import asyncio
import functools
import logging
import warnings

from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from traceback import format_exception
from typing import Any, Callable, Coroutine, Optional, Union

# custom components
from .cron import CronSchedule


# async def ... function types
CoroFunction = Callable[[], Coroutine[Any, Any, Any]]
ErrorHandler = Callable[[str, Any, Exception], Coroutine[Any, Any, Any]]

default_error_handler = None  # target for the @on_error() decorator
trigger_registry = []  # target for start_triggers()


class BaseTrigger(ABC):
    """
    Abstract base class for all repeating trigger decorators. This class can only be inherited from,
    it cannot be instantiated. Any subclasses have to implement the :meth:`next_run` property

    Attributes
    ----------

    iter_args: Optional[:class:`list`]
        an optional list of arguments. The decorated function will be called once per list element,
        and the element will be passed to the decorated function as the first positional argument

    on_startup: Optional[:class:`bool`]
        whether to trigger a run of the decorated function on startup. Defaults to `True`

    autostart: Optional[:class:`bool`]
        whether to automatically start the trigger. Auto-starting it may cause required components to not
        have fully loaded and initialized. If you choose to disable autostart (which is the default),
        you can use `coc.ext.triggers.start_triggers()` to manually kick the trigger execution off once you
        have loaded all required resources

    error_handler: Optional[:class:`coc.ext.triggers.CoroFunction`]
        an optional function that will be called on each error incurred during the trigger execution

    logger: Optional[:class:`logging.Logger`]
        an optional logger instance implementing the logging.Logger functionality. Debug and error logs
        about the trigger execution will be logged to this logger

    loop: Optional[:class:`asyncio.AbstractEventLoop`]
        an optional event loop that the trigger execution will be appended to. If no loop is provided,
        the trigger will provision one using `asyncio.get_event_loop()`

    kwargs:
        any additional keyword arguments that will be passed to the decorated function every time it is called

    """

    def __init__(self,
                 *,  # disable positional arguments
                 iter_args: Optional[list] = None,
                 on_startup: Optional[bool] = True,
                 autostart: Optional[bool] = False,
                 error_handler: Optional[ErrorHandler] = None,
                 logger: Optional[logging.Logger] = None,
                 loop: Optional[asyncio.AbstractEventLoop] = None,
                 **kwargs):

        if not error_handler and not default_error_handler and not logger:
            warnings.warn(
                'No logger or error handler are defined. Without either of these components, any errors '
                'raised during the trigger executions will be silently ignored. If you declared a global '
                'error handler using the `@on_error()` decorator, you can safely ignore this warning or '
                'remove it entirely by placing the handler declaration before the trigger declarations',
                category=RuntimeWarning
            )

        self.iter_args = iter_args
        self.on_startup = on_startup
        self.autostart = autostart
        self.error_handler = error_handler
        self.logger = logger
        self.loop = loop or asyncio.get_event_loop()
        self.kwargs = kwargs

        self.task = None  # placeholder for the repeat task created in self.__wrapper
        self.__next_run_time = None  # placeholder for the current target run time

    def __call__(self, func: CoroFunction):
        return self.__wrapper(func)

    def __wrapper(self, func: CoroFunction):
        """The main workhorse. Handles the repetition of the decorated function
        as well as all logging and error handling
        """

        # fill any passed kwargs
        fixture = functools.partial(func, **self.kwargs)

        @functools.wraps(fixture)
        async def wrapped() -> None:
            async def inner():
                # maybe wait for next trigger cycle
                if not self.on_startup:
                    self.__next_run_time = self.next_run
                    if self.logger:
                        self.logger.info(
                            f'`on_startup` is set to `False`. First run of {self.__class__.__name__} for '
                            f'{func.__name__}: {self.__next_run_time.isoformat()}'
                        )
                    await self.sleep_until(self.__next_run_time)

                # repeat indefinitely
                while True:
                    if self.logger:
                        self.logger.info(f'Running {self.__class__.__name__} for {func.__name__}')

                    # safeguard against early triggers - apparently, a desync is possible
                    now = datetime.now().astimezone()
                    if self.__next_run_time is not None and now < self.__next_run_time:
                        await asyncio.sleep(max((self.__next_run_time - now).total_seconds(), 0))

                    # call the decorated function
                    try:
                        if self.iter_args:
                            results = await asyncio.gather(*map(fixture, self.iter_args), return_exceptions=True)

                            # check for exceptions
                            for arg, res in zip(self.iter_args, results):
                                if isinstance(res, Exception):
                                    await self.__handle_exception(func, arg, res)
                        else:
                            await fixture()
                    except Exception as e:
                        await self.__handle_exception(func, None, e)

                    # sleep until next execution time
                    self.__next_run_time = self.next_run
                    if self.logger and datetime.now().astimezone() <= self.__next_run_time:
                        self.logger.info(
                            f'{self.__class__.__name__} finished for {func.__name__}. '
                            f'Next run: {self.__next_run_time.isoformat()}'
                        )
                    elif self.logger:  # i.e. next_run is in the past
                        self.logger.warning(
                            f'{self.__class__.__name__} missed the scheduled run time for {func.__name__}. Running now'
                        )

                    await self.sleep_until(self.__next_run_time)

            # create a reference to the repeating task to prevent it from accidentally being garbage collected
            self.task = self.loop.create_task(inner())

        if self.autostart:  # immediately start the trigger
            if self.logger:
                self.logger.info(f'{self.__class__.__name__} for {func.__name__} auto-started')
            self.loop.create_task(wrapped())
        else:  # add trigger to registry
            trigger_registry.append(wrapped())
            if self.logger:
                self.logger.info(f'{self.__class__.__name__} for {func.__name__} registered for manual start')

        return wrapped

    async def __handle_exception(self, func: CoroFunction, arg: Any, exc: Exception):
        """Handle exceptions during trigger calls. This will attempt to call the provided logger and
        any available error handler
        """

        if self.logger:
            self.logger.error(f'function: {func.__name__}, failing iter_arg: {arg}\n'
                              ''.join(format_exception(type(exc), exc, exc.__traceback__)))

        error_handler = self.error_handler or default_error_handler
        if error_handler:
            await error_handler(func.__name__, arg, exc)

    @staticmethod
    async def sleep_until(wakeup_date: datetime):
        """Sleep until a defined point in time. If that point is in the past, don't sleep at all
        Parameters
        ----------
        wakeup_date: :class:`datetime.datetime`
            a timezone-aware datetime at which the trigger should wake up again
        """

        await asyncio.sleep(max((wakeup_date - datetime.now().astimezone()).total_seconds(), 0))

    @property
    @abstractmethod
    def next_run(self) -> datetime:
        """Calculate the date and time of the next run. Needs to be overwritten in subclasses"""
        raise NotImplementedError('All `BaseTrigger` subclasses need to implement `next_run`')


class IntervalTrigger(BaseTrigger):
    """
    A decorator class to repeat a function every `seconds` seconds after the previous execution finishes

    Attributes
    ----------

    seconds: :class:`int`
        how many seconds to wait between trigger runs

    iter_args: Optional[:class:`list`]
        an optional list of arguments. The decorated function will be called once per list element,
        and the element will be passed to the decorated function as the first positional argument. If
        no iter_args are defined, nothing (especially not `None`) will be injected into the decorated function

    on_startup: Optional[:class:`bool`]
        whether to trigger a run of the decorated function on startup. Defaults to `True`

    autostart: Optional[:class:`bool`]
        whether to automatically start the trigger. Auto-starting it may cause required components to not
        have fully loaded and initialized. If you choose to disable autostart (which is the default),
        you can use `coc.ext.triggers.start_triggers()` to manually kick the trigger execution off once you
        have loaded all required resources

    error_handler: Optional[:class:`coc.ext.triggers.CoroFunction`]
        an optional coroutine function that will be called on each error incurred during the trigger execution.
        The handler will receive three arguments:

            function_name: :class:`str`
                the name of the failing trigger's decorated function
            arg: Optional[:class:`Any`]
                the failing `iter_args` element or None if no iter_args are defined
            exception: :class:`Exception`
                the exception that occurred

    logger: Optional[:class:`logging.Logger`]
        an optional logger instance implementing the logging.Logger functionality. Debug, warning and error logs
        about the trigger execution will be sent to this logger

    loop: Optional[:class:`asyncio.AbstractEventLoop`]
        an optional event loop that the trigger execution will be appended to. If no loop is provided,
        the trigger will provision one using `asyncio.get_event_loop()`

    kwargs:
        any additional keyword arguments that will be passed to the decorated function every time it is called

    Example
    -------
    .. code-block:: python3

        @IntervalTrigger(seconds=600, iter_args=['#2PP', '#2PPP'])
        async def download_current_war(clan_tag: str):
            # use your coc client to fetch war data, store it to a file or database, ...
            pass

    """

    def __init__(self,
                 *,  # disable positional arguments
                 seconds: int,
                 iter_args: Optional[list] = None,
                 on_startup: bool = True,
                 autostart: bool = False,
                 error_handler: Optional[CoroFunction] = None,
                 logger: Optional[logging.Logger] = None,
                 loop: Optional[asyncio.AbstractEventLoop] = None,
                 **kwargs):

        super().__init__(iter_args=iter_args, on_startup=on_startup, autostart=autostart,
                         error_handler=error_handler, logger=logger, loop=loop, **kwargs)

        if not isinstance(seconds, int) or seconds <= 0:
            raise ValueError(f'`seconds` must be a positive integer, got {seconds}')
        self._interval_seconds = seconds

    def __str__(self):
        return f'coc.ext.triggers.IntervalTrigger(seconds={self._interval_seconds})'

    @property
    def next_run(self) -> datetime:
        """Calculate the date and time of the next run based on the current time and the defined interval

        Returns
        -------
        the next run date (timezone-aware): :class:`datetime.datetime`
        """

        return datetime.now().astimezone() + timedelta(seconds=self._interval_seconds)

    @classmethod
    def hourly(cls, iter_args: Optional[list] = None, on_startup: bool = True, autostart: bool = False,
               error_handler: Optional[CoroFunction] = None, logger: Optional[logging.Logger] = None,
               loop: Optional[asyncio.AbstractEventLoop] = None, **kwargs):
        """A shortcut to create a trigger that runs with a one-hour break between executions"""

        return cls(seconds=3600, iter_args=iter_args, on_startup=on_startup, autostart=autostart,
                   error_handler=error_handler, logger=logger, loop=loop, **kwargs)

    @classmethod
    def daily(cls, iter_args: Optional[list] = None, on_startup: bool = True, autostart: bool = False,
              error_handler: Optional[CoroFunction] = None, logger: Optional[logging.Logger] = None,
              loop: Optional[asyncio.AbstractEventLoop] = None, **kwargs):
        """A shortcut to create a trigger that runs with a 24-hour break between executions"""

        return cls(seconds=86400, iter_args=iter_args, on_startup=on_startup, autostart=autostart,
                   error_handler=error_handler, logger=logger, loop=loop, **kwargs)


class CronTrigger(BaseTrigger):
    """
    A decorator class to repeat a function based on a Cron schedule

    Attributes
    ----------
    cron_schedule: Union[:class:`str`, :class:`coc.ext.triggers.CronSchedule`]
        the Cron schedule to follow

    iter_args: Optional[:class:`list`]
        an optional list of arguments. The decorated function will be called once per list element,
        and the element will be passed to the decorated function as the first positional argument. If
        no iter_args are defined, nothing (especially not `None`) will be injected into the decorated function

    on_startup: Optional[:class:`bool`]
        whether to trigger a run of the decorated function on startup. Defaults to `True`

    autostart: Optional[:class:`bool`]
        whether to automatically start the trigger. Auto-starting it may cause required components to not
        have fully loaded and initialized. If you choose to disable autostart (which is the default),
        you can use `coc.ext.triggers.start_triggers()` to manually kick the trigger execution off once you
        have loaded all required resources

    error_handler: Optional[:class:`coc.ext.triggers.CoroFunction`]
        an optional coroutine function that will be called on each error incurred during the trigger execution.
        The handler will receive three arguments:

            function_name: :class:`str`
                the name of the failing trigger's decorated function
            arg: Optional[:class:`Any`]
                the failing `iter_args` element or None if no iter_args are defined
            exception: :class:`Exception`
                the exception that occurred

    logger: Optional[:class:`logging.Logger`]
        an optional logger instance implementing the logging.Logger functionality. Debug, warning and error logs
        about the trigger execution will be sent to this logger

    loop: Optional[:class:`asyncio.AbstractEventLoop`]
        an optional event loop that the trigger execution will be appended to. If no loop is provided,
        the trigger will provision one using `asyncio.get_event_loop()`

    kwargs:
        any additional keyword arguments that will be passed to the decorated function every time it is called


    Example
    -------

    .. code-block:: python3

        @CronTrigger(cron_schedule='0 0 * * *', iter_args=['#2PP', '#2PPP'])
        async def download_current_war(clan_tag: str):
            # use your coc client to fetch war data, store it to a file or database,
            pass

    """

    def __init__(self,
                 *,  # disable positional arguments
                 cron_schedule: Union[CronSchedule, str],
                 iter_args: Optional[list] = None,
                 on_startup: bool = True,
                 autostart: bool = False,
                 error_handler: Optional[CoroFunction] = None,
                 logger: Optional[logging.Logger] = None,
                 loop: Optional[asyncio.AbstractEventLoop] = None,
                 **kwargs):

        super().__init__(iter_args=iter_args, on_startup=on_startup, autostart=autostart,
                         error_handler=error_handler, logger=logger, loop=loop, **kwargs)

        if isinstance(cron_schedule, str):
            cron_schedule = CronSchedule(cron_schedule)
        self.cron_schedule = cron_schedule

    def __str__(self):
        return f'coc.ext.triggers.CronTrigger(cron_schedule="{self.cron_schedule.cron_str}")'

    @property
    def next_run(self) -> datetime:
        """Calculate the date and time of the next run based on the current time and the defined Cron schedule

        Returns
        -------
        :class:`datetime.datetime`
            the next run date (timezone-aware):
        """

        # prevent multiple runs in one minute
        now = datetime.now().astimezone()
        return self.cron_schedule.next_run_after(now.replace(second=0, microsecond=0) + timedelta(minutes=1))

    @classmethod
    def hourly(cls, iter_args: Optional[list] = None, on_startup: bool = True, autostart: bool = False,
               error_handler: Optional[CoroFunction] = None, logger: Optional[logging.Logger] = None,
               loop: Optional[asyncio.AbstractEventLoop] = None, **kwargs):
        """A shortcut to create a trigger that runs at the start of every hour"""

        return cls(cron_schedule='0 * * * *', iter_args=iter_args, on_startup=on_startup, autostart=autostart,
                   error_handler=error_handler, logger=logger, loop=loop, **kwargs)

    @classmethod
    def daily(cls, iter_args: Optional[list] = None, on_startup: bool = True, autostart: bool = False,
              error_handler: Optional[CoroFunction] = None, logger: Optional[logging.Logger] = None,
              loop: Optional[asyncio.AbstractEventLoop] = None, **kwargs):
        """A shortcut to create a trigger that runs at the start of every day"""

        return cls(cron_schedule='0 0 * * *', iter_args=iter_args, on_startup=on_startup, autostart=autostart,
                   error_handler=error_handler, logger=logger, loop=loop, **kwargs)

    @classmethod
    def weekly(cls, iter_args: Optional[list] = None, on_startup: bool = True, autostart: bool = False,
               error_handler: Optional[CoroFunction] = None, logger: Optional[logging.Logger] = None,
               loop: Optional[asyncio.AbstractEventLoop] = None, **kwargs):
        """A shortcut to create a trigger that runs at the start of every week (Sunday at 00:00)"""

        return cls(cron_schedule='0 0 * * 0', iter_args=iter_args, on_startup=on_startup, autostart=autostart,
                   error_handler=error_handler, logger=logger, loop=loop, **kwargs)

    @classmethod
    def monthly(cls, iter_args: Optional[list] = None, on_startup: bool = True, autostart: bool = False,
                error_handler: Optional[CoroFunction] = None, logger: Optional[logging.Logger] = None,
                loop: Optional[asyncio.AbstractEventLoop] = None, **kwargs):
        """A shortcut to create a trigger that runs at the start of every month"""

        return cls(cron_schedule='0 0 1 * *', iter_args=iter_args, on_startup=on_startup, autostart=autostart,
                   error_handler=error_handler, logger=logger, loop=loop, **kwargs)


def on_error() -> Callable[[], ErrorHandler]:
    """A decorator function that designates a function as the global fallback error handler for all exceptions
    during trigger executions.

    Notes
    -----
    This handler declaration should occur before any trigger declarations to avoid a RuntimeWarning about a
    potentially undeclared error handler, though that warning can safely be ignored.

    Any function decorated by this must be a coroutine and accept three parameters:

        function_name: :class:`str`
            the name of the failing trigger's decorated function
        arg: Optional[:class:`Any`]
            the failing `iter_args` element or None if no iter_args are defined
        exception: :class:`Exception`
            the exception that occurred

    Returns
    -------
    the decorated handler function

    Example
    --------
    .. code-block:: python3

        @on_error()
        async def handle_trigger_exception(function_name: str, arg: Any, exception: Exception):
            # send a Discord message, do some data cleanup, ...
            pass

    """

    def wrapper(func: ErrorHandler):
        # register the error handler
        global default_error_handler
        default_error_handler = func

        @functools.wraps(func)
        async def wrapped(function_name, arg, error):
            await func(function_name, arg, error)
        return wrapped
    return wrapper


async def start_triggers():
    """Manually start all triggers with `autostart=False` (which is the default value)

    Example
    --------
    .. code-block:: python3

        # define a trigger
        @CronTrigger(cron_schedule='0 0 * * *', iter_args=['#2PP', '#2PPP'], autostart=False)
        async def download_current_war(clan_tag: str):
            # use your coc client to fetch war data, store it to a file or database, ...
            pass

        if __name__ = '__main__':
            # login to clashy.py and/or load other required resources here
            event_loop = asyncio.get_event_loop()
            # then start trigger execution
            event_loop.run_until_complete(start_triggers())
            # set the loop to run forever so that it keeps executing the triggers
            event_loop.run_forever()
    """

    tasks = [asyncio.create_task(trigger) for trigger in trigger_registry]
    return await asyncio.gather(*tasks)



import json
import unittest
from pathlib import Path

import orjson

from coc import RaidLogEntry, RaidClan, RaidAttack, RaidMember
from coc.entry_logs import RaidLog

import tracemalloc

tracemalloc.start()

with open(Path(__file__).parent.joinpath(Path("mockdata/clans/capitalraidseasons/CAPITALRAIDSEASON.json")),'rb') as fp:
	MOCK_CAPITALRAIDSEASON = orjson.loads(fp.read())


class TestRaids(unittest.TestCase):
	def test_raid_log(self):
		data = MOCK_CAPITALRAIDSEASON["body"]
		clan_tag = MOCK_CAPITALRAIDSEASON["clantag"]
		raidlog = RaidLog(client=None, clan_tag=clan_tag, limit=2,
						  page=False, json_resp=data, model=RaidLogEntry,
                          lookup_cache=False, update_cache=False, ignore_cached_errors=[])
		# test last raid log entry
		self.raidlog_entry(raidlog[0], data["items"][0])

		# test second last raid log entry
		self.raidlog_entry(raidlog[1], data["items"][1])

	def raidlog_entry(self, raidlogentry, mock_data):
		map_raw_to_cocpy = {
			"state"                  : "state",
			"capitalTotalLoot"       : "total_loot",
			"raidsCompleted"         : "completed_raid_count",
			"totalAttacks"           : "attack_count",
			"enemyDistrictsDestroyed": "destroyed_district_count",
			"offensiveReward"        : "offensive_reward",
			"defensiveReward"        : "defensive_reward"
		}
		for k, v in map_raw_to_cocpy.items():
			self.assertEqual(mock_data.get(k), raidlogentry.__getattribute__(v),
							 f'{k=} {v=} {mock_data.get(k)=} {raidlogentry.__getattribute__(v)}')

		# test non trivial data

		# test start time
		self.assertEqual(mock_data.get("startTime"), raidlogentry.start_time.raw_time)

		# test end time
		self.assertEqual(mock_data.get("endTime"), raidlogentry.end_time.raw_time)

		# test attack log
		for c, adata in enumerate(mock_data.get("attackLog", [])):
			self.raid_clan(raidlogentry.attack_log[c], adata, c)

		# test defense log
		for c, adata in enumerate(mock_data.get("defenseLog", [])):
			self.raid_clan(raidlogentry.defense_log[c], adata, c)

		# test members
		for member in mock_data.get("members", []):
			self.raid_member(raidlogentry, member)

			# test member attacks
			mock_attacks = [attack for attack_raid in mock_data.get("attackLog", []) for district in attack_raid.get(
					"districts", []) for attack in district.get("attacks", [])
							if attack and attack.get("attacker", {}).get("tag") == member.get("tag")]
			raidmember = raidlogentry.get_member(member.get("tag"))
			attacks = raidmember.attacks
			self.assertEqual(len(mock_attacks), len(attacks))
			self.assertEqual(len(mock_attacks), raidmember.attack_count)
			for attack, mock_attack in zip(attacks, mock_attacks):
				self.raid_attack(attack, mock_attack)

		# test defensive custom properties
		defensive_loot = raidlogentry.total_defensive_loot
		defensive_attack_count = raidlogentry.defense_attack_count
		defensive_destroyed_districts = raidlogentry.defensive_destroyed_district_count
		defensive_loot_mock = 0
		defensive_attack_count_mock = 0
		defensive_destroyed_districts_mock = 0
		for clan_data in mock_data.get("defenseLog", []):
			defensive_destroyed_districts_mock += clan_data.get("districtsDestroyed", 0)
			defensive_attack_count_mock += clan_data.get("attackCount", 0)
			for district_data in clan_data.get("districts", []):
				defensive_loot_mock += district_data.get("totalLooted", 0)
		self.assertEqual(defensive_loot, defensive_loot_mock, "RaidLogEntry.total_defensive_loot")
		self.assertEqual(defensive_attack_count, defensive_attack_count_mock, "RaidLogEntry.defense_attack_count")
		self.assertEqual(defensive_destroyed_districts, defensive_destroyed_districts_mock,
						 "RaidLogEntry.defensive_destroyed_district_count")

	def raid_clan(self, raidclan: RaidClan, mock_data, index):
		map_raw_to_cocpy = {
			"attackCount"       : "attack_count",
			"districtCount"     : "district_count",
			"districtsDestroyed": "destroyed_district_count"
		}
		self.assertEqual(raidclan.tag, mock_data.get("attacker", mock_data.get("defender")).get("tag"))
		self.assertEqual(raidclan.name, mock_data.get("attacker", mock_data.get("defender")).get("name"))
		self.assertEqual(raidclan.level, mock_data.get("attacker", mock_data.get("defender")).get("level"))
		self.assertEqual(raidclan.index, index)
		for k, v in map_raw_to_cocpy.items():
			self.assertEqual(mock_data.get(k), raidclan.__getattribute__(v),
							 f'{k=} {v=} {mock_data.get(k)=} {raidclan.__getattribute__(v)}')

		# test non trivial data

		# test raid districts
		loot_mock = 0
		destroyed_mock = 0
		for c, data in enumerate(mock_data.get("districts", [])):
			self.raid_district(raidclan.districts[c], data)
			loot_mock += data.get("totalLooted", 0)
			destroyed_mock += 1 if data.get("destructionPercent") == 100 else 0

		self.assertEqual(loot_mock, raidclan.looted, "looted")
		self.assertEqual(destroyed_mock == mock_data.get("districtCount"), raidclan.is_finished, "is_finished")

		# test raid clan attacks
		mock_attacks = [attack for district in mock_data.get("districts", []) for attack in district.get("attacks", [])]
		attacks = raidclan.attacks
		self.assertEqual(len(mock_attacks), len(attacks))
		for data, mock in zip(attacks, mock_attacks):
			self.raid_attack(data, mock)

	def raid_district(self, raiddistrict, mock_data):
		map_raw_to_cocpy = {
			"id"                : "id",
			"name"              : "name",
			"districtHallLevel" : "hall_level",
			"destructionPercent": "destruction",
			"attackCount"       : "attack_count",
			"totalLooted"       : "looted"
		}

		for k, v in map_raw_to_cocpy.items():
			self.assertEqual(mock_data.get(k), raiddistrict.__getattribute__(v),
							 f'{k=} {v=} {mock_data.get(k)=} {raiddistrict.__getattribute__(v)}')

		# test non trivial data

		# test raid districts
		for c, data in enumerate(mock_data.get("attacks", [])):
			self.raid_attack(raiddistrict.attacks[c], data)

	def raid_attack(self, raidattack: RaidAttack, mock_data):
		self.assertEqual(mock_data.get("attacker", mock_data.get("defender", {})).get("tag"), raidattack.attacker_tag)
		self.assertEqual(mock_data.get("attacker", mock_data.get("defender", {})).get("name"), raidattack.attacker_name)
		self.assertEqual(mock_data.get("stars"), raidattack.stars)
		self.assertEqual(mock_data.get("destructionPercent"), raidattack.destruction)

	def raid_member(self, raidlogentry: RaidLogEntry, mock_data):
		raidmember = raidlogentry.get_member(mock_data.get("tag"))
		self.assertIsInstance(raidmember, RaidMember)
		self.assertEqual(raidmember.name, mock_data.get("name"))
		self.assertEqual(raidmember.attack_count, mock_data.get("attacks"))
		self.assertEqual(raidmember.attack_limit, mock_data.get("attackLimit"))
		self.assertEqual(raidmember.bonus_attack_limit, mock_data.get("bonusAttackLimit"))
		self.assertEqual(raidmember.capital_resources_looted, mock_data.get("capitalResourcesLooted"))




import asyncio
import logging
from enum import Enum

from itertools import cycle
from pathlib import Path
from typing import AsyncIterator, Iterable, List, Optional, Type, Union, TYPE_CHECKING

import orjson

from .game_data import AccountData, ArmyRecipe, StaticData
from .clans import Clan, RankedClan
from .errors import Forbidden, GatewayError, NotFound, PrivateWarLog
from .enums import WarRound
from .miscmodels import BaseLeague, GoldPassSeason, Label, League, Location, LoadGameData, Translation
from .http import HTTPClient, BasicThrottler, BatchThrottler
from .iterators import (
    PlayerIterator,
    ClanIterator,
    ClanWarIterator,
    LeagueWarIterator,
    CurrentWarIterator,
    SeasonIterator
)
from .players import Player, ClanMember, RankedPlayer
from .hero import Hero, Pet, Equipment
from .spell import Spell
from .troop import Troop
from .raid import RaidLogEntry
from .utils import correct_tag, get
from .wars import ClanWar, ClanWarLogEntry, ClanWarLeagueGroup, ExtendedCWLGroup
from .entry_logs import ClanWarLog, RaidLog


LOG = logging.getLogger(__name__)

LEAGUE_WAR_STATE = "notInWar"
KEY_MINIMUM, KEY_MAXIMUM = 1, 10

TRANSLATION_PATH = Path(__file__).parent.joinpath(Path("static/translations.json"))
STATIC_DATA_PATH = Path(__file__).parent.joinpath(Path("static/static_data.json"))

class ClashAccountScopes(Enum):
    """
    Values represent the scope required for each type of user. A USER is
    anyone who has access to the API. A REAL user is a user with special
    access from SuperCell with realtime scope access.
    """
    USER = "clash"
    REAL = "clash:*:verifytoken,realtime"


class Client:
    """This is the client connection used to interact with the Clash of Clans API.

    Parameters
    ----------
    key_count : int
        The amount of keys to use for this client. Maximum of 10.
        Defaults to 1.

    key_names : str
        Default name for keys created to use for this client.
        All keys created or to be used with this client must
        have this name.
        Defaults to "Created with clashy.py Client".

    throttle_limit : int
        The number of requests per token per second to send to the API.
        Once hitting this limit, the library will automatically throttle
        your requests.

        .. note::

            Setting this value too high may result in the API rate-limiting
            your requests. This means you cannot request for ~30-60 seconds.

        .. warning::

            Setting this value too high may result in your requests being
            deemed "API Abuse", potentially resulting in an IP ban.

        Defaults to 10 requests per token, per second.

    loop : :class:`asyncio.AbstractEventLoop`, optional
        The :class:`asyncio.AbstractEventLoop` to use for HTTP requests.
        An :func:`asyncio.get_event_loop()` will be used if ``None`` is passed

    correct_tags : :class:`bool`
        Whether the client should correct tags before requesting them from the API.
        This process involves stripping tags of whitespace and adding a `#` prefix if not present.
        Defaults to ``True``.

    connector : :class:`aiohttp.BaseConnector`
        The aiohttp connector to use. By default, this is ``None``.

    timeout: :class:`float`
        The number of seconds before timing out with an API query. Defaults to 30.

    cache_max_size: :class:`int`
        The max size of the internal cache layer. Defaults to 10 000. Set this to ``None`` to remove any cache layer.

    load_game_data: :class:`LoadGameData`
        The option for how clashy.py will load game data. See :ref:`initialising_game_data` for more info.

    realtime: :class:`bool`
        Some developers are given special access to an uncached API access by
        Super Cell. If you are one of those developers, your account will have
        special flags that will only be interpreted by clashy.py if you set this
        bool to True.

    raw_attribute: :class:`bool`
        The option to enable the _raw_data attribute for most objects in the library. This attribute will contain
        the original json data as returned by the API. This can be useful if you want to store a response in a database
        for later use or are interested in new things that clashy.py does not support otherwise yet. But because this
        increases the memory footprint and is not needed for most use cases, this defaults to ``False``.

    base_url: :class:`str`
        The base URL to use for API requests. Defaults to "https://api.clashofclans.com/v1"

    ip: :class:`str`
        The IP address to use for API requests. Defaults to None, which means the IP address will be automatically
        detected.

    lookup_cache: :class:`bool`
        Flag for controlling the cache usage before an actual API request is made. Defaults to True, which means the cache lookup is done

    update_cache: :class:`bool`
        Flag for controlling if the cache is updated after an API request was made. Defaults to True, which means the cache is updated
        after an API request.

    ignore_cached_errors: :class:`list[int]`
        In case of a cache lookup and a cached entry exists, ignore the cached data if the status code of the response is in the list.

    player_cls: :class:`Type[Player]`
        Class to be used for player objects. Defaults to :class:`Player`.

    member_cls: :class:`Type[ClanMember]`
        Class to be used for clan member objects. Defaults to :class:`ClanMember`.

    ranke

    clan_cls: :class:`Type[Clan]`
        Class to be used for clan objects. Defaults to :class:`Clan`.


    Attributes
    ----------
    loop : :class:`asyncio.AbstractEventLoop`
        The loop that is used for HTTP requests
    static_data : :class:`StaticData`
        The full static game data loaded from game files
    """

    __slots__ = (
        "base_url",
        "ip",
        "loop",
        "correct_key_count",
        "key_names",
        "key_scopes",
        "throttle_limit",
        "throttler",
        "timeout",
        "connector",
        "cache_max_size",
        "stats_max_size",
        "http",
        "realtime",
        "raw_attribute",
        "_ready",
        "correct_tags",
        "load_game_data",
        "lookup_cache",
        "update_cache",
        "ignore_cached_errors",
        "_players",
        "_clans",
        "_wars",
        "objects_cls",
        "_troop_holder",
        "_spell_holder",
        "_hero_holder",
        "_pet_holder",
        "_equipment_holder",
        "_static_data",
        "_translations",
        "_name_to_id_mapping",
        "static_data",
    )

    def __init__(
        self,
        *,
        key_count: int = 1,
        key_names: str = "Created with clashy.py Client",
        throttle_limit: int = 30,
        loop: asyncio.AbstractEventLoop = None,
        correct_tags: bool = True,
        throttler: Type[Union[BasicThrottler, BatchThrottler]] = BasicThrottler,
        connector=None,
        timeout: float = 30.0,
        cache_max_size: int = 10000,
        stats_max_size: int = 1000,
        load_game_data: LoadGameData = LoadGameData(default=True),
        realtime=False,
        raw_attribute=False,
        base_url: str = "https://api.clashofclans.com/v1",
        ip: Optional[str] = None,
        lookup_cache: Optional[bool] = True,
        update_cache: Optional[bool] = True,
        ignore_cached_errors: Union[List[int], None] = None,
        **kwargs,
    ):

        self.loop = loop or asyncio.get_event_loop()

        self.correct_key_count = max(min(KEY_MAXIMUM, key_count), KEY_MINIMUM)

        if not key_count == self.correct_key_count:
            raise RuntimeError("Key count must be within {}-{}".format(KEY_MINIMUM, KEY_MAXIMUM))

        self.key_names = key_names
        self.key_scopes = ClashAccountScopes.REAL.value if realtime else ClashAccountScopes.USER.value
        self.throttle_limit = throttle_limit
        self.throttler = throttler
        self.connector = connector
        self.timeout = timeout
        self.cache_max_size = cache_max_size
        self.stats_max_size = stats_max_size

        self.lookup_cache = lookup_cache
        self.update_cache = update_cache
        self.ignore_cached_errors = ignore_cached_errors

        self.http: Optional[HTTPClient] = None  # set in method login()
        self.realtime = realtime
        self.raw_attribute = raw_attribute
        self.correct_tags = correct_tags
        self.load_game_data = load_game_data
        self.base_url = base_url
        self.ip = ip

        self.objects_cls = {"Player": Player, "Clan": Clan, "ClanWar": ClanWar,
                            "RankedPlayer": RankedPlayer, "RankedClan": RankedClan,
                            "ClanMember": ClanMember, "ClanWarLogEntry": ClanWarLogEntry, "RaidLogEntry": RaidLogEntry,
                            "ClanWarLeagueGroup": ClanWarLeagueGroup, "Location": Location,
                            "League": League, "BaseLeague": BaseLeague, "GoldPassSeason": GoldPassSeason,
                            "Label": Label}

        # cache
        self._players = {}
        self._clans = {}
        self._wars = {}

        self._translations = {}
        self._static_data = {}
        self._name_to_id_mapping = {}
        self.static_data: StaticData = ...

    @property
    def _defaults(self):
        return {
            "lookup_cache": self.lookup_cache,
            "update_cache": self.update_cache,
            "ignore_cached_errors": self.ignore_cached_errors,
            "realtime": self.realtime,
        }

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self.close()

    def set_object_cls(self, name: str, cls):
        """Set a custom class type for a given object name. The method ensures that the
        provided class is a valid subclass of a predefined default class. It updates
        the internal mapping of object classes to the new custom class.

        .. note::

            This affects only the return type of Client methods returning the object.
            For example changing the ClanMember class to a custom class will affect 'get_clan_members', but not the clan members
            of a clan object obtained by calling `get_clan` or `get_clans` methods.

        Parameters
        -----------
        name: str
            The name of the object whose class type is to be set. It must be one of
            the predefined object types such as "Player", "Clan", etc.
        cls:
            The custom class type to be associated with the given object name. The
            class must be a subclass of the corresponding default class type.

        Raises
        -------
        ValueError
            If the provided object name is not supported.
        TypeError
            If the provided class is not a subclass of the default class for the given
            object name.
        """
        default_cls = {"Player": Player, "Clan": Clan, "ClanWar": ClanWar,
                       "RankedPlayer": RankedPlayer, "RankedClan": RankedClan,
                       "ClanMember": ClanMember, "ClanWarLogEntry": ClanWarLogEntry, "RaidLogEntry": RaidLogEntry,
                       "ClanWarLeagueGroup": ClanWarLeagueGroup, "Location": Location,
                       "League": League, "BaseLeague": BaseLeague, "GoldPassSeason": GoldPassSeason,
                       "Label": Label}
        if name not in default_cls:
            raise ValueError(f"Setting a cls with the name {name} is not supported.")
        if not issubclass(cls, default_cls[name]):
            raise TypeError(f"The cls {cls} must be a subclass of {default_cls[name]}")
        self.objects_cls[name] = cls

    def _create_client(self, email, password):
        return HTTPClient(
            client=self,
            email=email,
            password=password,
            key_names=self.key_names,
            key_scopes=self.key_scopes,
            loop=self.loop,
            key_count=self.correct_key_count,
            throttle_limit=self.throttle_limit,
            throttler=self.throttler,
            cache_max_size=self.cache_max_size,
            stats_max_size=self.stats_max_size,
            base_url=self.base_url,
            ip=self.ip,
            lookup_cache=self.lookup_cache,
            update_cache=self.update_cache,
            ignore_cached_errors=self.ignore_cached_errors,
        )

    def _load_static(self):
        with open(TRANSLATION_PATH, 'rb') as fp:
            self._translations = orjson.loads(fp.read())

        with open(STATIC_DATA_PATH, 'rb') as fp:
            static_data = orjson.loads(fp.read())
            self.static_data = StaticData(data=static_data)
            id_mapped_static_data = {}

            for section, items in static_data.items():
                for item in items:
                    if "_id" not in item: # skips over achievements data, since we don't have ids for those
                        continue
                    id_mapped_static_data[item['_id']] = item

                    if section == "troops":
                        self._name_to_id_mapping[(item['name'], section, item.get("village"))] = item['_id']
                    else:
                        self._name_to_id_mapping[(item['name'], section)] = item['_id']

            self._static_data = id_mapped_static_data

    def _get_static_data(self,
        item_name: str = None,
        section: str = None,
        village: str | None = None,
        item_id: str = None,
        bypass: bool = False
    ):
        if bypass:
            return None
        if item_id:
            return self._static_data.get(item_id)

        if village:
            item_id = self._name_to_id_mapping.get((item_name, section, village))
            return self._static_data.get(item_id)

        item_id = self._name_to_id_mapping.get((item_name, section))
        return self._static_data.get(item_id)

    async def login(self, email: str, password: str) -> None:
        """Retrieves all keys and creates an HTTP connection ready for use.

        Parameters
        ----------
        email : str
            Your password email from https://developer.clashofclans.com
            This is used when updating keys automatically if your IP changes

        password : str
            Your password login from https://developer.clashofclans.com
            This is used when updating keys automatically if your IP changes
        """
        self.http = http = self._create_client(email, password)
        await http.create_session(self.connector, self.timeout)
        await http.initialise_keys()

        self._load_static()
        LOG.debug("HTTP connection created. Client is ready for use.")

    def login_with_keys(self, *keys: str) -> None:
        """Creates an HTTP connection ready for use with the keys you provide.

        .. deprecated:: v2.3.0
            This function has been deemed deprecated to allow
            asyncio to clean up the async structures. Please use :func:`Client.login_with_tokens`
            instead.

        Parameters
        ----------
        keys: list[str]
            Keys or tokens as found from https://developer.clashofclans.com.


        """
        self.correct_key_count = len(keys)
        self.http = http = self._create_client(None, None)
        http._keys = keys
        http.keys = cycle(http._keys)
        self.loop.run_until_complete(http.create_session(self.connector, self.timeout))
        self._load_static()

        LOG.debug("HTTP connection created. Client is ready for use.")

    async def login_with_tokens(self, *tokens: str) -> None:
        """Creates an HTTP connection ready for use with the tokens you provide.

        Parameters
        ----------
        tokens: list[str]
            Tokens as found from https://developer.clashofclans.com under "My account" -> <your key> -> "token".
        """
        self.correct_key_count = len(tokens)
        self.http = http = self._create_client(None, None)
        http._keys = tokens
        http.keys = cycle(http._keys)
        await http.create_session(self.connector, self.timeout)
        self._load_static()

        LOG.debug("HTTP connection created. Client is ready for use.")

    async def close(self) -> None:
        """Closes the HTTP connection from within a loop function such as
        async def main()"""
        await self.http.close()

    def dispatch(self, event_name: str, *args, **kwargs) -> None:
        """Dispatches an event listener matching the `event_name` parameter."""
        LOG.debug("Dispatching %s event", event_name)

        try:
            fctn = getattr(self, event_name)
        except AttributeError:
            return

        if asyncio.iscoroutinefunction(fctn):
            self.loop.create_task(fctn(*args, **kwargs))
        else:
            fctn(*args, **kwargs)

    async def search_clans(
        self,
        *,
        name: str = None,
        war_frequency: str = None,
        location_id: int = None,
        min_members: int = None,
        max_members: int = None,
        min_clan_points: int = None,
        min_clan_level: int = None,
        label_ids: List[Union[Label, int]] = [],
        limit: int = None,
        before: str = None,
        after: str = None,
        cls: Type[Clan] = None,
        **kwargs,
    ) -> List[Clan]:
        """Search all clans by name and/or filtering the results using various criteria.

        At least one filtering criteria must be defined and if name is used as part
        of search, it is required to be at least three characters long.

        Parameters
        -----------
        name : str, optional
            The clan name.
        war_frequency : str, optional
            The war frequency.
        location_id : int, optional
            The location id.
        min_members : int, optional
            The minimum number of members.
        max_members : int, optional
            The maximum number of members.
        min_clan_points : int, optional
            The minumum clan points.
        min_clan_level : int, optional
            The minimum clan level.
        label_ids: :class:`List`[Union[:class:`coc.Label`, :class:`int`]]
            List of Labels or Label ids
        limit : int
            The number of clans to search for.
        cls:
            Target class to use to model that data returned

        Raises
        -------
        RuntimeError
            At least one filtering parameter must be passed.
        TypeError
            The ``cls`` parameter must be a subclass of :class:`Clan`.
        Maintenance
            The API is currently in maintenance.
        GatewayError
            The API hit an unexpected gateway exception.

        Returns
        --------
        List[:class:`Clan`]
            A list of all clans found matching criteria provided.

        """
        if not (
            name or war_frequency or location_id or min_members or max_members or min_clan_points or min_clan_level
        ):
            raise RuntimeError("At least one filtering parameter must be passed.")
        if cls is None:
            cls = self.objects_cls['Clan']
        if not issubclass(cls, Clan):
            raise TypeError("cls must be a subclass of Clan.")

        data = await self.http.search_clans(
            name=name,
            warFrequency=war_frequency,
            locationId=location_id,
            minMembers=min_members,
            maxMembers=max_members,
            minClanPoints=min_clan_points,
            minClanLevel=min_clan_level,
            label_ids=",".join([str(x.id) if isinstance(x, Label) else str(x) for x in label_ids
                                if isinstance(x, (Label, int,))]),
            limit=limit,
            before=before,
            after=after,
            **{**self._defaults, **kwargs}
        )
        return [cls(data=n, client=self, **kwargs) for n in data.get("items", [])]


    async def get_clan(self, tag: str, cls: Type[Clan] = None, **kwargs) -> Clan:
        """Get information about a single clan by clan tag.

        Clan tags can be found using clan search operation.

        Parameters
        -----------
        tag : str
            The clan tag to search for.

        cls:
            Target class to use to model that data returned

        Raises
        -------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`Clan`.

        NotFound
            No clan was found with the supplied tag.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        :class:`Clan`
            The clan with provided tag.
        """
        if cls is None:
            cls = self.objects_cls['Clan']
        if not issubclass(cls, Clan):
            raise TypeError("cls must be a subclass of Clan.")

        if self.correct_tags:
            tag = correct_tag(tag)

        data = await self.http.get_clan(tag,
                                        lookup_cache=kwargs.get("lookup_cache", self.lookup_cache),
                                        update_cache=kwargs.get("update_cache", self.update_cache),
                                        ignore_cached_errors=kwargs.get("ignore_cached_errors", self.ignore_cached_errors))
        return cls(data=data, client=self, **kwargs)

    def get_clans(self, tags: Iterable[str], cls: Type[Clan] = None, **kwargs) -> AsyncIterator[Clan]:
        """Get information about multiple clans by clan tag.
        Refer to `Client.get_clan` for more information.

        This returns a :class:`ClanIterator` which fetches the requested clan tags in parallel.

        Example
        ---------

        .. code-block:: python3

            tags = [...]
            async for clan in client.get_clans(tags):
                print(clan.name)

        Parameters
        -----------
        tags : Iterable[:class:`str`]
            An iterable of clan tags to search for.

        cls:
            Target class to use to model that data returned

        Raises
        ------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`Clan`.


        Yields
        ------
        :class:`Clan`
            A clan matching one of the tags requested.
        """
        if cls is None:
            cls = self.objects_cls['Clan']
        if not issubclass(cls, Clan):
            raise TypeError("cls must be a subclass of Clan.")

        return ClanIterator(self, tags, cls, **{**self._defaults, **kwargs})


    async def get_members(self, clan_tag: str, *, limit: int = 0, after: str = "", before: str = "",
                          cls: Type[ClanMember] = None, **kwargs) -> List[ClanMember]:
        """List clan members.

        This is equivilant to ``(await Client.get_clan('tag')).members``.

        Parameters
        -----------
        clan_tag : str
            The clan tag to search for.
        cls:
            Target class to use to model that data returned
        limit:
            class:`int`: Number of members to retrieve

        after:
            class:`str`: Pagination string to get page after

        before:
            class:`str`: Pagination string to get page before


        Raises
        -------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`ClanMember`.

        NotFound
            No clan was found with the supplied tag.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        List[:class:`ClanMember`]
            A list of members in the clan.
        """
        if cls is None:
            cls = self.objects_cls['ClanMember']
        if not issubclass(cls, ClanMember):
            raise TypeError("cls must be a subclass of ClanMember.")

        if self.correct_tags:
            clan_tag = correct_tag(clan_tag)

        args = {}
        if limit:
            args['limit'] = limit
        if after:
            args['after'] = after
        if before:
            args['before'] = before
        args['lookup_cache'] = kwargs.get("lookup_cache", self.lookup_cache)
        args['update_cache'] = kwargs.get("update_cache", self.update_cache)
        args['ignore_cached_errors'] = kwargs.get("ignore_cached_errors", self.ignore_cached_errors)

        data = await self.http.get_clan_members(clan_tag, **args)
        return [cls(data=mdata, client=self, **kwargs) for mdata in data.get("items", [])]

    async def get_war_log(
        self,
        clan_tag: str,
        cls: Type[ClanWarLogEntry] = None,
        page: bool = False,
        *,
        limit: int = 0,
        after: str = "",
        before: str = "",
        **kwargs
    ) -> ClanWarLog:
        """
        Retrieve a clan's clan war log. By default, this will return
        all the clan's log available in the API. This will of course consume
        memory. The option of limiting the amount of log items fetched
        can be controlled with the `limit` parameter. Additionally, if
        `paginate` is set to True, and an async for loop is performed
        on this object, then additional log items will be fetched but only
        consume the same amount of memory space at all time.


        .. note::

            Please see documentation for :class:`ClanWarLogEntry` for different attributes
            which are present when the entry is a regular clan war or a league clan war.
            The difference can be found with :attr:`ClanWarLogEntry.is_league_entry`.


        Parameters
        -----------
        clan_tag:
            class:`str`: The clan tag to search for.

        cls:
            Target class to use to model that data returned

        page:
            class:`bool`: Enable fetching logs while only holding the
            same amount of logs as `limit`. If `paginate` is set to True,
            and `limit` is set to default of 0, then `limit` will be set to
            10 automatically.

        limit:
            class:`int`: Number of logs to retrieve

        after:
            class:`str`: Pagination string to get page after

        before:
            class:`str`: Pagination string to get page before

        Raises
        ------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`ClanWarLogEntry`.

        NotFound
            No clan was found with the supplied tag.

        PrivateWarLog
            The clan's warlog is private.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        :class:`ClanWarLog`:
            Entries in the warlog of the requested clan.
        """
        if limit < 0:
            raise ValueError("Limit cannot be negative")
        if cls is None:
            cls = self.objects_cls['ClanWarLogEntry']
        if not issubclass(cls, ClanWarLogEntry):
            raise TypeError("cls must be a subclass of ClanWarLogEntry.")

        if self.correct_tags:
            clan_tag = correct_tag(clan_tag)

        # If paginate is enabled and limit is set to default of 0, then
        # set limit to a new default of 10
        if page:
            limit = limit if limit else 10



        try:
            return await ClanWarLog.init_cls(client=self,
                                             clan_tag=clan_tag,
                                             page=page,
                                             limit=limit,
                                             model=cls,
                                             after=after,
                                             before=before,
                                             **{**self._defaults, **kwargs})
        except Forbidden as exception:
            raise PrivateWarLog(exception.response,
                                exception.reason) from exception

    async def get_raid_log(
            self,
            clan_tag: str,
            cls: Type[RaidLogEntry] = None,
            page: bool = False,
            *,
            limit: int = 0,
            after: str = "",
            before: str = "",
            **kwargs
    ) -> RaidLog:
        """
        Retrieve a clan's Capital Raid Log. By default, this will return
        all the clan's log available in the API. This will of course consume
        memory. The option of limiting the amount of log items fetched
        can be controlled with the `limit` parameter. Additionally, if
        `paginate` is set to True, and an async for loop is performed
        on this object, then additional log items will be fetched but only
        consume the same amount of memory space at all time.


        Parameters
        -----------
        clan_tag:
            class:`str`: The clan tag to search for.

        cls:
            Target class to use to model that data returned

        page:
            class:`bool`: Enable fetching logs while only holding the
            same amount of logs as `limit`. If `paginate` is set to True,
            and `limit` is set to default of 0, then `limit` will be set to
            10 automatically.

        limit:
            class:`int`: Number of logs to retrieve

        after:
            class:`str`: Pagination string to get page after

        before:
            class:`str`: Pagination string to get page before

        Raises
        ------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`RaidLogEntry`.

        NotFound
            No clan was found with the supplied tag.

        PrivateWarLog
            The clan's warlog is private.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        :class:`RaidLog`:
            Entries in the capital raid seasons of the requested clan.
        """

        if limit < 0:
            raise ValueError("Limit cannot be negative")
        if cls is None:
            cls = self.objects_cls['RaidLogEntry']
        if not issubclass(cls, RaidLogEntry):
            raise TypeError("cls must be a subclass of ClanWarLogEntry.")

        if self.correct_tags:
            clan_tag = correct_tag(clan_tag)

        # If paginate is enabled and limit is set to default of 0, then
        # set limit to a new default of 10
        if page:
            limit = limit if limit else 10

        try:
            return await RaidLog.init_cls(client=self,
                                          clan_tag=clan_tag,
                                          page=page,
                                          limit=limit,
                                          model=cls,
                                          after=after,
                                          before=before,
                                          **{**self._defaults, **kwargs}
                                          )
        except Forbidden as exception:
            raise PrivateWarLog(exception.response,
                                exception.reason) from exception

    async def get_clan_war(self, clan_tag: str, cls: Type[ClanWar] = None, **kwargs) -> ClanWar:
        """
        Retrieve information about clan's current clan war

        Parameters
        -----------
        clan_tag : str
            The clan tag to search for.

        Raises
        ------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`ClanWar`.

        NotFound
            No clan was found with the supplied tag.

        PrivateWarLog
            The clan's war log is private.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        :class:`ClanWar`
            The clan's current war.

        """
        if cls is None:
            cls = self.objects_cls['ClanWar']
        if not issubclass(cls, ClanWar):
            raise TypeError("cls must be a subclass of ClanWar.")

        if self.correct_tags:
            clan_tag = correct_tag(clan_tag)

        try:
            data = await self.http.get_clan_current_war(clan_tag, **{**self._defaults, **kwargs})
        except Forbidden as exception:
            raise PrivateWarLog(exception.response, exception.reason) from exception

        return cls(data=data, client=self, clan_tag=clan_tag, **kwargs)

    def get_clan_wars(self, clan_tags: Iterable[str], cls: Type[ClanWar] = None, **kwargs) -> AsyncIterator[ClanWar]:
        """
        Retrieve information multiple clan's current clan wars

        This returns a :class:`coc.WarIterator` which fetches the requested wars in parallel.

        .. note ::

            This will skip any clans who have a private war-log.


        .. note ::

            This will not fetch any CWL wars. Use :meth:`Client.get_current_wars` for that.


        Example
        ---------

        .. code-block:: python3

            tags = [...]
            async for clan_war in Client.get_clan_wars(tags):
                print(clan_war.opponent)

        Parameters
        -----------
        clan_tags : Iterable[:class:`str`]
            An iterable of clan tags to search for.

        cls:
            Target class to use to model that data returned

        Raises
        ------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`ClanWar`.

        NotFound
            No clan was found with the supplied tag.

        PrivateWarLog
            The clan's warlog is private.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Yields
        ------
        :class:`ClanWar`
            A war matching one of the tags requested.
        """
        if cls is None:
            cls = self.objects_cls['ClanWar']
        if not issubclass(cls, ClanWar):
            raise TypeError("cls must be a subclass of ClanWar.")

        return ClanWarIterator(self, clan_tags, cls=cls, **{**self._defaults, **kwargs})

    async def get_league_group(
        self,
        clan_tag: str,
        cls: Type[ClanWarLeagueGroup] = None,
        **kwargs
    ) -> ClanWarLeagueGroup:
        """Retrieve information about clan's current clan war league group.

        Parameters
        -----------
        clan_tag : str
            The clan tag to search for.

        cls:
            Target class to use to model that data returned

        Raises
        ------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`ClanWarLeagueGroup`.

        NotFound
            No clan was found with the supplied tag.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.
            This may be due to an API bug whereby requesting
            the league group of a clan searching for a CWL match will time out.


        Returns
        --------
        :class:`ClanWarLeagueGroup`
            The clan's war league group.
        """
        if cls is None:
            cls = self.objects_cls['ClanWarLeagueGroup']
        # pylint: disable=protected-access
        if not issubclass(cls, ClanWarLeagueGroup):
            raise TypeError("cls must be a subclass of ClanWarLeagueGroup.")

        if self.correct_tags:
            clan_tag = correct_tag(clan_tag)

        try:
            data = await self.http.get_clan_war_league_group(clan_tag, **{**self._defaults, **kwargs})
        except Forbidden as exception:
            raise PrivateWarLog(exception.response, exception.reason) from exception
        except asyncio.TimeoutError:
            raise GatewayError(
                "Client timed out waiting for %s clan tag. This may be the result of an API bug which times out "
                "when requesting the league group of a clan searching for a Clan War League match."
            )

        return cls(data=data, client=self, **kwargs)

    async def get_league_war(self, war_tag: str, cls: Type[ClanWar] = None, **kwargs) -> ClanWar:
        """
        Retrieve information about a clan war league war.

        Parameters
        -----------
        war_tag : str
            The league war tag to search for.

        cls:
            Target class to use to model that data returned

        Raises
        ------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`ClanWar`.

        NotFound
            No clan was found with the supplied tag.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        :class:`ClanWar`
            The league war associated with the war tag
        """
        if cls is None:
            cls = self.objects_cls['ClanWar']
        # pylint: disable=protected-access
        if not issubclass(cls, ClanWar):
            raise TypeError("cls must be a subclass of LeagueWar.")

        if self.correct_tags:
            war_tag = correct_tag(war_tag)

        try:
            data = await self.http.get_cwl_wars(war_tag, **{**self._defaults, **kwargs})
        except Forbidden as exception:
            raise PrivateWarLog(exception.response, exception.reason) from exception

        data["tag"] = war_tag  # API doesn't return this, even though it is in docs.
        return cls(data=data, client=self, **kwargs)

    def get_league_wars(
        self,
        war_tags: Iterable[str],
        clan_tag: str = None,
        cls: Type[ClanWar] = None,
        **kwargs
    ) -> AsyncIterator[ClanWar]:
        """
        Retrieve information about multiple league wars

        This returns a :class:`LeagueWarIterator` which fetches the requested clan tags in parallel.

        Example
        ---------

        .. code-block:: python3

            tags = [...]
            async for league_war in Client.get_league_wars(tags):
                print(league_war.opponent)

        Parameters
        -----------
        war_tags : Iterable[:class:`str`]
            An iterable of war tags to search for.
        clan_tag: :class:`str`
            An optional clan tag. If present, this will only return wars which belong to this clan.
        cls:
            Target class to use to model that data returned

        Raises
        ------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`ClanWar`.

        Yields
        ------
        :class:`ClanWar`
            A war matching one of the tags requested.
        """
        if cls is None:
            cls = self.objects_cls['ClanWar']
        if not issubclass(cls, ClanWar):
            raise TypeError("cls must be a subclass of ClanWar.")

        return LeagueWarIterator(self, war_tags, clan_tag, cls, **{**self._defaults, **kwargs})

    async def get_current_war(
        self,
        clan_tag: str,
        cwl_round: WarRound = WarRound.current_war,
        cls: Type[ClanWar] = None,
        **kwargs
    ) -> Optional[ClanWar]:
        """Retrieve a clan's current war.

        Unlike ``Client.get_clan_war`` or ``Client.get_league_war``,
        this method will search for a regular war, and if the clan is in ``notInWar`` state,
        search for a current league war.

        This simplifies what would otherwise be 2-3 function calls to find a war.

        If you don't wish to search for CWL wars, use :meth:`Client.get_clan_war`.

        This method will consume the :exc:`PrivateWarLog` error, instead returning ``None``.

        .. note::

            You can differentiate between a regular clan war and a clan war league (CWL) war
            by using the helper property, :attr:`ClanWar.is_cwl`.


        Parameters
        -----------
        clan_tag : str
            An iterable of clan tag to search for.

        cwl_round: :class:`WarRound`
            An enum detailing the type of round to get. Could be ``coc.WarRound.previous_war``,
            ``coc.WarRound.current_war`` or ``coc.WarRound.preparation``.
            This defaults to ``coc.WarRound.current_war``.

        cls:
            Target class to use to model that data returned

        Raises
        -------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`ClanWar`.

        NotFound
            No clan was found with the supplied tag.

        PrivateWarLog
            The clan's warlog is private.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        :class:`ClanWar`
            The clan's current war. Could be ``None``.

            If the clan is in CWL, the league group can be accessed via :attr:`ClanWar.league_group`.

            If you pass in a :class:`WarRound` and that round doesn't exist (yet), this will return ``None``.
        """
        kwargs = {**self._defaults, **kwargs}
        if cls is None:
            cls = self.objects_cls['ClanWar']
        # pylint: disable=protected-access
        if not issubclass(cls, ClanWar):
            raise TypeError("cls must be a subclass of ClanWar.")

        if self.correct_tags:
            clan_tag = correct_tag(clan_tag)

        try:
            get_war = await self.get_clan_war(clan_tag, cls=cls, **kwargs)
        except PrivateWarLog:
            get_war = None

        if get_war and get_war.state != LEAGUE_WAR_STATE:
            return get_war

        try:
            league_group = await self.get_league_group(clan_tag, **kwargs)
        except (NotFound, GatewayError) as exception:
            # either they're not in cwl (NotFound)
            # or it's an API bug where league group endpoint will timeout when the clan is searching (GatewayError)
            if get_war is None:
                raise PrivateWarLog(exception.response, exception.reason) from exception
            return get_war

        if league_group.state == "notInWar" or league_group.state == "groupNotFound":
            return None
        last_round_active = league_group.number_of_rounds == len(league_group.rounds)
        if last_round_active and league_group.state != "ended":
            # there are the supposed number of rounds, but without any call we are unable to know if the last round is
            # currently in preparation or already in war
            async for war in self.get_league_wars(league_group.rounds[-1], cls=cls, **kwargs):
                if war.state == 'inWar':
                    # last round is already in war
                    last_round_active = True
                    break
                elif war.state == 'preparation':
                    # last round is still in preparation
                    last_round_active = False
                    break
        if cwl_round is WarRound.current_war and league_group.state == "preparation":
            return None  # for round 1 and 15min prep between rounds this is a shortcut.
        elif cwl_round is WarRound.current_preparation and league_group.state == "ended":
            return None  # for the end of CWL there's no next prep day.
        elif cwl_round is WarRound.current_war and len(league_group.rounds) < 2:
            round_tags = league_group.rounds[-1] # for the first round during prep already return round 1
        elif cwl_round is WarRound.current_war and (last_round_active or league_group.state == "ended"):
            round_tags = league_group.rounds[-1]  # for the end of CWL current_war should give the last war
        elif cwl_round is WarRound.previous_war and (last_round_active or league_group.state == "ended"):
            round_tags = league_group.rounds[-2]  # for the end of CWL previous_war should give the second last war
        elif cwl_round is WarRound.previous_war and len(league_group.rounds) < 3:
            return None  # no previous war for two rounds.
        elif cwl_round is WarRound.previous_war:
            round_tags = league_group.rounds[-3]
        elif cwl_round is WarRound.current_war:
            round_tags = league_group.rounds[-2]
        elif cwl_round is WarRound.current_preparation:
            round_tags = league_group.rounds[-1]
        else:
            return None

        kwargs["league_group"] = league_group
        kwargs["clan_tag"] = clan_tag
        async for war in self.get_league_wars(round_tags, cls=cls, **kwargs):
            if war.clan_tag == clan_tag:
                return war
            elif war.opponent.tag == clan_tag:
                tmp = war.clan
                war.clan = war.opponent
                war.opponent = tmp
                return war

    def get_current_wars(
        self,
        clan_tags: Iterable[str],
        cls: Type[ClanWar] = None,
        **kwargs
    ) -> AsyncIterator[ClanWar]:
        """Retrieve information multiple clan's current wars.

        See :meth:`Client.get_current_war` for more information.

        This returns a :class:`CurrentWarIterator` which fetches the requested clan tags in parallel.

        .. note ::

            This will skip any clans who have a private war-log.


        .. note ::

            This will fetch CWL wars, which may result in a slower operation due to multiple API calls.
            If you only wish to get regular wars, consider :meth:`Client.get_clan_wars`.


        Example
        ---------

        .. code-block:: python3

            tags = [...]
            async for war in Client.get_current_wars(tags):
                print(war.type)

        Parameters
        -----------
        clan_tags : Iterable[:class:`str`]
            An iterable of clan tags to search for.

        cls:
            Target class to use to model that data returned

        Raises
        ------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`ClanWar`.

        Yields
        --------
        :class:`ClanWar`
            The clan war of a requested tag.
        """
        if not isinstance(clan_tags, Iterable):
            raise TypeError("Tags are not an iterable.")
        if cls is None:
            cls = self.objects_cls.get("ClanWar")
        if not issubclass(cls, ClanWar):
            raise TypeError("cls must be a subclass of ClanWar.")

        return CurrentWarIterator(client=self, tags=clan_tags, cls=cls, **{**self._defaults, **kwargs})

    async def search_locations(self, *, limit: int = None, before: str = None, after: str = None, cls: Type[Location] = None,
                               **kwargs) -> List[Location]:
        """List all available locations

        Parameters
        -----------
        limit : int, optional
            Number of items to fetch. Default is None, which returns all available locations
        before : str, optional
            For use with paging. Not implemented yet.
        after: str, optional
            For use with paging. Not implemented yet.
        cls:
            Target class to use to model that data returned

        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        List[:class:`Location`]
            The requested locations.
        """
        if cls is None:
            cls = self.objects_cls['Location']
        if not issubclass(cls, Location):
            raise TypeError("cls must be a subclass of Location.")
        data = await self.http.search_locations(limit=limit, before=before, after=after, **{**self._defaults, **kwargs})

        return [cls(data=n) for n in data["items"]]

    async def get_location(self, location_id: int, cls: Type[Location] = None, **kwargs) -> Location:
        """Get information about specific location

        Parameters
        -----------
        location_id : int
            The Location ID to search for.
        cls:
            Target class to use to model that data returned

        Raises
        ------
        NotFound
            No location was found with the supplied ID.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        :class:`Location`
            The requested location.
        """
        if cls is None:
            cls = self.objects_cls['Location']
        if not issubclass(cls, Location):
            raise TypeError("cls must be a subclass of Location.")
        data = await self.http.get_location(location_id, **{**self._defaults, **kwargs})
        return cls(data=data)

    async def get_location_named(self, location_name: str, cls: Type[Location] = None, **kwargs) -> Optional[Location]:
        """Get a location by name.

        This is equivalent to:

        .. code-block:: python3

            locations = await client.search_locations(limit=None)
            return utils.get(locations, name=location_name)


        Parameters
        -----------
        location_name : str
            The location name to search for
        cls:
            Target class to use to model that data returned

        Returns
        --------
        :class:`Location`
            The first location matching the location name.
        """
        if cls is None:
            cls = self.objects_cls['Location']
        if not issubclass(cls, Location):
            raise TypeError("cls must be a subclass of Location.")
        data = await self.http.search_locations(limit=None, before=None, after=None, **{**self._defaults, **kwargs})
        locations = [cls(data=n) for n in data["items"]]

        return get(locations, name=location_name)

    async def get_location_clans(
            self, location_id: int = "global", *, limit: int = None,
            before: str = None, after: str = None, cls: Type[RankedClan] = None,
            **kwargs
    ) -> List[RankedClan]:
        """Get clan rankings for a specific location

        Parameters
        -----------
        location_id : int
            The Location ID to search for. Defaults to all locations (``global``).
        limit : int
            The number of results to fetch.
        before : str, optional
            For use with paging. Not implemented yet.
        after: str, optional
            For use with paging. Not implemented yet.
        cls:
            Target class to use to model that data returned

        Raises
        ------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`RankedClan`.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        List[:class:`RankedClan`]
            The top clans for the requested location.
        """
        if cls is None:
            cls = self.objects_cls['RankedClan']
        if not issubclass(cls, RankedClan):
            raise TypeError("cls must be a subclass of RankedClan.")
        data = await self.http.get_location_clans(location_id, limit=limit, before=before, after=after,
                                                  **{**self._defaults, **kwargs})
        return [cls(data=n, client=self) for n in data["items"]]

    async def get_location_clans_capital(
            self, location_id: int = "global", *, limit: int = None,
            before: str = None, after: str = None, cls: Type[RankedClan] = None,
            **kwargs
    ) -> List[RankedClan]:
        """Get clan capital rankings for a specific location

        Parameters
        -----------
        location_id : int
            The Location ID to search for. Defaults to all locations (``global``).
        limit : int
            The number of results to fetch.
        before : str, optional
            For use with paging. Not implemented yet.
        after: str, optional
            For use with paging. Not implemented yet.
        cls:
            Target class to use to model that data returned

        Raises
        ------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`RankedClan`.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        List[:class:`RankedClan`]
            The top clans for the requested location.
        """
        if cls is None:
            cls = self.objects_cls['RankedClan']
        if not issubclass(cls, RankedClan):
            raise TypeError("cls must be a subclass of RankedClan.")
        data = await self.http.get_location_clans_capital(location_id, limit=limit, before=before, after=after,
                                                          **{**self._defaults, **kwargs})
        return [cls(data=n, client=self) for n in data["items"]]

    async def get_location_players(
            self, location_id: int = "global", *, limit: int = None,
            before: str = None, after: str = None, cls: Type[RankedPlayer] = None,
            **kwargs
    ) -> List[RankedPlayer]:
        """Get player rankings for a specific location

        Parameters
        -----------
        location_id : int
            The Location ID to search for. Defaults to all locations (global).
        limit : int
            The number of results to fetch.
        before : str, optional
            For use with paging. Not implemented yet.
        after: str, optional
            For use with paging. Not implemented yet.
        cls:
            Target class to use to model that data returned

        Raises
        ------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`RankedPlayer`.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        List[:class:`RankedPlayer`]
            The top players for the requested location.
        """
        if cls is None:
            cls = self.objects_cls['RankedPlayer']
        if not issubclass(cls, RankedPlayer):
            raise TypeError("cls must be a subclass of RankedPlayer.")
        data = await self.http.get_location_players(location_id, limit=limit, before=before, after=after,
                                                    **{**self._defaults, **kwargs})
        return [cls(data=n, client=self) for n in data["items"]]

    async def get_location_clans_builder_base(
            self, location_id: int = "global", *, limit: int = None,
            before: str = None, after: str = None, cls: Type[RankedClan] = None, **kwargs
    ) -> List[RankedClan]:
        """Get clan builder base rankings for a specific location

        Parameters
        -----------
        location_id : int
            The Location ID to search for. Defaults to all locations (global).
        limit : int
            The number of results to fetch.
        before : str, optional
            For use with paging. Not implemented yet.
        after: str, optional
            For use with paging. Not implemented yet.
        cls:
            Target class to use to model that data returned

        Raises
        ------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`RankedClan`.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        List[:class:`RankedClan`]
            The top builder base-clans for the requested location.
        """
        if cls is None:
            cls = self.objects_cls['RankedClan']
        if not issubclass(cls, RankedClan):
            raise TypeError("cls must be a subclass of RankedClan.")
        data = await self.http.get_location_clans_builder_base(location_id, limit=limit, before=before, after=after,
                                                               **{**self._defaults, **kwargs})
        return [cls(data=n, client=self) for n in data["items"]]

    async def get_location_players_builder_base(
            self, location_id: int = "global", *, limit: int = None,
            before: str = None, after: str = None, cls: Type[RankedPlayer] = None, **kwargs
    ) -> List[RankedPlayer]:
        """Get player builder base rankings for a specific location

        Parameters
        -----------
        location_id : int
            The Location ID to search for. Defaults to all locations (global).
        limit : int
            The number of results to fetch.
        before : str, optional
            For use with paging. Not implemented yet.
        after: str, optional
            For use with paging. Not implemented yet.
        cls:
            Target class to use to model that data returned

        Raises
        ------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`RankedPlayer`.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        List[:class:`RankedPlayer`]
            The top builder base players for the requested location.
        """
        if cls is None:
            cls = self.objects_cls['RankedPlayer']
        if not issubclass(cls, RankedPlayer):
            raise TypeError("cls must be a subclass of RankedPlayer.")
        data = await self.http.get_location_players_builder_base(location_id, limit=limit, before=before, after=after,
                                                                 **{**self._defaults, **kwargs})
        return [cls(data=n, client=self) for n in data["items"]]

    # leagues
    async def search_leagues(self, *, limit: int = None, before: str = None, after: str = None, cls: Type[League] = None,
                             **kwargs) -> List[League]:
        """Get list of leagues.

        Parameters
        -----------
        limit : int
            Number of items to fetch. Defaults to ``None`` (all leagues).
        before : str, optional
            For use with paging. Not implemented yet.
        after: str, optional
            For use with paging. Not implemented yet.


        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        List[:class:`League`]
            The requested leagues.
        """
        if cls is None:
            cls = self.objects_cls['League']
        if not issubclass(cls, League):
            raise TypeError("cls must be a subclass of League.")
        data = await self.http.search_leagues(limit=limit, before=before, after=after, **{**self._defaults, **kwargs})
        return [cls(data=n, client=self) for n in data["items"]]

    async def get_league(self, league_id: int, cls: Type[League] = None, **kwargs) -> League:
        """
        Get league information

        Parameters
        -----------
        league_id : str
            The League ID to search for.

        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.

        NotFound
            No league was found with the supplied league ID.


        Returns
        --------
        :class:`League`
            The league with the requested ID
        """
        if cls is None:
            cls = self.objects_cls['League']
        if not issubclass(cls, League):
            raise TypeError("cls must be a subclass of League.")
        data = await self.http.get_league(league_id, **{**self._defaults, **kwargs})
        return cls(data=data, client=self)

    async def get_league_named(self, league_name: str, cls: Type[League] = None, **kwargs) -> Optional[League]:
        """Get a league by name.

        This is somewhat equivalent to

        .. code-block:: python3

            leagues = await client.search_leagues(limit=None)
            return utils.get(leagues, name=league_name)


        Parameters
        -----------
        league_name : str
            The league name to search for
        cls:
            Target class to use to model that data returned

        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.

        Returns
        --------
        :class:`League`
            The first league matching the league name. Could be ``None`` if not found.
        """
        if cls is None:
            cls = self.objects_cls['League']
        if not issubclass(cls, League):
            raise TypeError("cls must be a subclass of League.")
        return get(await self.search_leagues(cls=cls, **{**self._defaults, **kwargs}), name=league_name)

    async def search_builder_base_leagues(self, *, limit: int = None, before: str = None, after: str = None,
                                          cls: Type[BaseLeague] = None, **kwargs)-> List[BaseLeague]:
        """Get list of builder base leagues.

        Parameters
        -----------
        limit : int
            Number of items to fetch. Defaults to ``None`` (all leagues).
        before : str, optional
            For use with paging. Not implemented yet.
        after: str, optional
            For use with paging. Not implemented yet.
        cls:
            Target class to use to model that data returned.


        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        List[:class:`BaseLeague`]
            The requested leagues.
        """
        if cls is None:
            cls = self.objects_cls['BaseLeague']
        if not issubclass(cls, BaseLeague):
            raise TypeError("cls must be a subclass of BaseLeague.")
        data = await self.http.search_builder_base_leagues(limit=limit, before=before, after=after,
                                                           **{**self._defaults, **kwargs})
        return [cls(data=n, client=self) for n in data["items"]]

    async def get_builder_base_league(self, league_id: int, cls: Type[BaseLeague] = None, **kwargs) -> BaseLeague:
        """
        Get builder base league information

        Parameters
        -----------
        league_id : str
            The League ID to search for.
        cls:
            Target class to use to model that data returned.

        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.

        NotFound
            No league was found with the supplied league ID.


        Returns
        --------
        :class:`BaseLeague`
            The league with the requested ID
        """
        if cls is None:
            cls = self.objects_cls['BaseLeague']
        if not issubclass(cls, BaseLeague):
            raise TypeError("cls must be a subclass of BaseLeague.")
        data = await self.http.get_builder_base_league(league_id, **{**self._defaults, **kwargs})
        return cls(data=data, client=self)

    async def get_builder_base_league_named(self, league_name: str, cls: Type[BaseLeague] = None, **kwargs) -> Optional[BaseLeague]:
        """Get a builder base league by name.

        This is somewhat equivalent to

        .. code-block:: python3

            leagues = await client.search_builder_base_leagues(limit=None)
            return utils.get(leagues, name=league_name)


        Parameters
        -----------
        league_name : str
            The builder base league name to search for
        cls:
            Target class to use to model that data returned.

        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.

        Returns
        --------
        :class:`BaseLeague`
            The first league matching the league name. Could be ``None`` if not found.
        """
        return get(await self.search_builder_base_leagues(cls=cls, **{**self._defaults, **kwargs}), name=league_name)

    async def search_war_leagues(self, *, limit: int = None, before: str = None, after: str = None, cls: Type[BaseLeague] = None,
                                 **kwargs) -> List[BaseLeague]:
        """Get list of war leagues.

        Parameters
        -----------
        limit : int
            Number of items to fetch. Defaults to ``None`` (all leagues).
        before : str, optional
            For use with paging. Not implemented yet.
        after: str, optional
            For use with paging. Not implemented yet.
        cls:
            Target class to use to model that data returned.


        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        List[:class:`BaseLeague`]
            The requested leagues.
        """
        if cls is None:
            cls = self.objects_cls['BaseLeague']
        if not issubclass(cls, BaseLeague):
            raise TypeError("cls must be a subclass of BaseLeague.")
        data = await self.http.search_war_leagues(limit=limit, before=before, after=after, **{**self._defaults, **kwargs})
        return [cls(data=n, client=self) for n in data["items"]]

    async def get_war_league(self, league_id: int, cls: Type[BaseLeague] = None, **kwargs) -> BaseLeague:
        """
        Get war league information

        Parameters
        -----------
        league_id : str
            The League ID to search for.
        cls:
            Target class to use to model that data returned.

        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.

        NotFound
            No league was found with the supplied league ID.


        Returns
        --------
        :class:`BaseLeague`
            The league with the requested ID
        """
        if cls is None:
            cls = self.objects_cls['BaseLeague']
        if not issubclass(cls, BaseLeague):
            raise TypeError("cls must be a subclass of BaseLeague.")
        data = await self.http.get_war_league(league_id, **{**self._defaults, **kwargs})
        return cls(data=data, client=self)

    async def get_war_league_named(self, league_name: str, cls: Type[BaseLeague] = None, **kwargs) -> Optional[BaseLeague]:
        """Get a war league by name.

        This is somewhat equivalent to

        .. code-block:: python3

            leagues = await client.search_war_leagues(limit=None)
            return utils.get(leagues, name=league_name)


        Parameters
        -----------
        league_name : str
            The war league name to search for
        cls:
            Target class to use to model that data returned.

        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.

        Returns
        --------
        :class:`BaseLeague`
            The first league matching the league name. Could be ``None`` if not found.
        """
        if cls is None:
            cls = self.objects_cls['BaseLeague']
        if not issubclass(cls, BaseLeague):
            raise TypeError("cls must be a subclass of BaseLeague.")
        return get(await self.search_war_leagues(cls=cls, **{**self._defaults, **kwargs}), name=league_name)

    async def search_capital_leagues(self, *, limit: int = None, before: str = None, after: str = None, cls: Type[BaseLeague] = None,
                                     **kwargs) -> List[BaseLeague]:
        """Get list of capital leagues.

        Parameters
        -----------
        limit : int
            Number of items to fetch. Defaults to ``None`` (all leagues).
        before : str, optional
            For use with paging. Not implemented yet.
        after: str, optional
            For use with paging. Not implemented yet.
        cls:
            Target class to use to model that data returned.

        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        List[:class:`BaseLeague`]
            The requested leagues.
        """
        if cls is None:
            cls = self.objects_cls['BaseLeague']
        if not issubclass(cls, BaseLeague):
            raise TypeError("cls must be a subclass of BaseLeague.")
        data = await self.http.search_capital_leagues(limit=limit, before=before, after=after,
                                                      **{**self._defaults, **kwargs})
        return [cls(data=n, client=self) for n in data["items"]]

    async def get_capital_league(self, league_id: int, cls: Type[BaseLeague] = None, **kwargs) -> BaseLeague:
        """
        Get capital league information

        Parameters
        -----------
        league_id : str
            The League ID to search for.
        cls:
            Target class to use to model that data returned.

        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.

        NotFound
            No league was found with the supplied league ID.


        Returns
        --------
        :class:`BaseLeague`
            The league with the requested ID
        """
        if cls is None:
            cls = self.objects_cls['BaseLeague']
        if not issubclass(cls, BaseLeague):
            raise TypeError("cls must be a subclass of BaseLeague.")
        data = await self.http.get_capital_league(league_id, **{**self._defaults, **kwargs})
        return cls(data=data, client=self)

    async def get_capital_league_named(self, league_name: str, cls: Type[BaseLeague] = None, **kwargs) -> Optional[BaseLeague]:
        """Get a capital league by name.

        This is somewhat equivalent to

        .. code-block:: python3

            leagues = await client.search_capital_leagues(limit=None)
            return utils.get(leagues, name=league_name)


        Parameters
        -----------
        league_name : str
            The capital league name to search for
        cls:
            Target class to use to model that data returned.

        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.

        Returns
        --------
        :class:`BaseLeague`
            The first league matching the league name. Could be ``None`` if not found.
        """
        if cls is None:
            cls = self.objects_cls['BaseLeague']
        if not issubclass(cls, BaseLeague):
            raise TypeError("cls must be a subclass of BaseLeague.")
        return get(await self.search_capital_leagues(cls=cls, **{**self._defaults, **kwargs}), name=league_name)

    async def get_seasons(self, league_id: int = 29000022, **kwargs) -> List[str]:
        """Get league seasons.

        .. note::

            League season information is available only for Legend League, with a league ID 29000022.


        Parameters
        -----------
        league_id : str
            The League ID to search for. Defaults to the only league you can get season information for, legends.
        cls:
            Target class to use to model that data returned.

        Raises
        ------
        InvalidArgument
            An invalid league_id was supplied. Currently, the only league supported is legends.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.

        Returns
        -------
        List[str]
            The legend season IDs, in the form ``YYYY-MM``, ie. ``2020-04``.
        """
        data = await self.http.get_league_seasons(league_id, **{**self._defaults, **kwargs})
        return [entry["id"] for entry in data["items"]]

    async def get_season_rankings(self, league_id: int, season_id: str, cls: Type[RankedPlayer] = None, **kwargs) -> AsyncIterator[RankedPlayer]:
        """Get league season rankings.

        .. note::

            League season information is available only for Legend League, with a league ID 29000022.


        Parameters
        -----------
        league_id : str
            The League ID to search for.
        season_id : str
            The Season ID to search for.
        cls:
            Target class to use to model that data returned.

        Raises
        ------
        InvalidArgument
            An invalid league_id or season_id was supplied. Currently, the only league supported is legends.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        List[:class:`RankedPlayer`]
            Top players for the requested season and league.
        """
        if cls is None:
            cls = self.objects_cls['RankedPlayer']
        if not issubclass(cls, RankedPlayer):
            raise TypeError("cls must be a subclass of RankedPlayer.")
        return SeasonIterator(client=self, league_id=league_id, season_id=season_id, cls=cls, **{**self._defaults, **kwargs})


    async def get_clan_labels(self, *, limit: int = None, before: str = None, after: str = None, cls: Type[Label] = None, **kwargs
                              ) -> List[Label]:
        """Fetch all possible clan labels.

        Parameters
        -----------
        limit : int
            The number of results to fetch.
        before : str, optional
            For use with paging. Not implemented yet.
        after: str, optional
            For use with paging. Not implemented yet.
        cls:
            Target class to use to model that data returned.#
        
        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        List[:class:`Label`]
            A list of all possible clan labels.
        """
        if cls is None:
            cls = self.objects_cls['Label']
        if not issubclass(cls, Label):
            raise TypeError("cls must be a subclass of Label.")
        data = await self.http.get_clan_labels(limit=limit, before=before, after=after, **{**self._defaults, **kwargs})
        return [cls(data=n, client=self) for n in data["items"]]

    async def get_player_labels(self, *, limit: int = None, before: str = None, after: str = None, cls: Type[Label] = None, **kwargs
                                ) -> List[Label]:
        """Fetch all possible player labels.

        Parameters
        -----------
        limit : int
            The number of results to fetch.
        before : str, optional
            For use with paging. Not implemented yet.
        after: str, optional
            For use with paging. Not implemented yet.
        cls:
            Target class to use to model that data returned.
        
        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        List[:class:`Label`]
            A list of all possible player labels.
        """
        if cls is None:
            cls = self.objects_cls['Label']
        if not issubclass(cls, Label):
            raise TypeError("cls must be a subclass of Label.")
        data = await self.http.get_player_labels(limit=limit, before=before, after=after, **{**self._defaults, **kwargs})
        return [cls(data=n, client=self) for n in data["items"]]

    # players
    async def get_player(self, player_tag: str, cls: Type[Player] = Player,
                         load_game_data: bool = None, **kwargs) -> Player:
        """Get information about a single player by player tag.
        Player tags can be found either in game or by from clan member lists.

        Parameters
        ----------
        player_tag : str
            The player tag to search for.

        cls:
            Target class to use to model that data returned

        Raises
        -------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`Player`.

        NotFound
            No player was found with the supplied tag.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        :class:`Player`
            The player with the tag.
        """
        if cls is None:
            cls = self.objects_cls['Player']
        if not issubclass(cls, Player):
            raise TypeError("cls must be a subclass of Player.")
        if load_game_data and not isinstance(load_game_data, bool):
            raise TypeError("load_game_data must be either True or False.")

        if self.correct_tags:
            player_tag = correct_tag(player_tag)

        data = await self.http.get_player(player_tag)
        return cls(data=data, client=self, load_game_data=load_game_data, **{**self._defaults, **kwargs})

    def get_players(self, player_tags: Iterable[str], cls: Type[Player] = None, load_game_data: bool = None, **kwargs) -> AsyncIterator[
        Player]:
        """Get information about a multiple players by player tag.
        Player tags can be found either in game or by from clan member lists.

        This returns a :class:`PlayerIterator` which fetches the requested player tags in parallel.

        Example
        ---------

        .. code-block:: python3

            tags = [...]
            async for player in Client.get_players(tags):
                print(player)

        Parameters
        ----------
        player_tags : Iterable[:class:`str`]
            An iterable of player tags to search for.

        cls:
            Target class to use to model that data returned

        Raises
        ------
        TypeError
            The ``cls`` parameter must be a subclass of :class:`Player`.

        Yields
        ------
        :class:`Player`
            A player matching one of the tags requested.
        """
        if cls is None:
            cls = self.objects_cls['Player']
        if not issubclass(cls, Player):
            raise TypeError("cls must be a subclass of Player.")
        if load_game_data and not isinstance(load_game_data, bool):
            raise TypeError("load_game_data must be either True or False.")

        return PlayerIterator(self, player_tags, cls=cls, load_game_data=load_game_data, **{**self._defaults, **kwargs} )

    async def verify_player_token(self, player_tag: str, token: str, **kwargs) -> bool:
        """Verify player API token that can be found from the game settings.

        This can be used to check that players own the game accounts they claim to own
        as they need to provide the one-time use API token that exists inside the game.

        Parameters
        ----------
        player_tag : str
            The player tag to verify.
        token : str
            The player's API token to verify.

        Raises
        ------
        NotFound
            No player was found with the supplied tag.

        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.


        Returns
        --------
        :class:`bool`
            Whether the player tag and API token were a valid match.
        """
        if self.correct_tags:
            player_tag = correct_tag(player_tag)

        data = await self.http.verify_player_token(player_tag, token, lookup_cache=False, update_cache=False,
                                                   ignore_cached_errors=kwargs.get('ignore_cached_errors', self.ignore_cached_errors))
        return data and data["status"] == "ok" or False

    async def get_current_goldpass_season(self, cls: Type[GoldPassSeason] = None, **kwargs) -> GoldPassSeason:
        """Get the current gold pass season
        
        Parameters
        ----------
        cls:
            Target class to use to model that data returned.

        Raises
        ------
        Maintenance
            The API is currently in maintenance.

        GatewayError
            The API hit an unexpected gateway exception.

        Returns
        --------
        :class:`GoldPassSeason`
            The gold pass season object of the current season"""
        if cls is None:
            cls = self.objects_cls['GoldPassSeason']
        if not issubclass(cls, GoldPassSeason):
            raise TypeError("cls must be a subclass of GoldPassSeason.")
        data = await self.http.get_current_goldpass_season(**{**self._defaults, **kwargs})
        return cls(data=data)

    def parse_army_link(self, link: str) -> ArmyRecipe:
        """Transform an army link from in-game into an ArmyRecipe object.

        Example
        -------

        .. code-block:: python3

            army = client.parse_army_link("https://link.clashofclans.com/en?action=CopyArmy&army=u10x0-2x3s1x9-3x2")

            for troop, quantity in army.troops:
                print("The user wants {} {}s. They each have {} DPS.".format(quantity, troop.name, troop.dps))

            for spell, quantity in army.spells:
                print("The user wants {} {}s.".format(quantity, spell.name))

            for hero_loadout in army.heroes_loadout:
                print("Hero: {}, Pet: {}, Equipment: {}".format(
                    hero_loadout.hero.name,
                    hero_loadout.pet.name if hero_loadout.pet else None,
                    [eq.name for eq in hero_loadout.equipment]
                ))


        Parameters
        ----------
        link: str
            The army share link, as copied from in-game.
            It should look something like: https://link.clashofclans.com/en?action=CopyArmy&army=u10x0-2x3s1x9-3x2

        Raises
        ------
        RuntimeError
            Troop and Spell game metadata must be loaded to use this feature.

        Returns
        --------
        :class:`ArmyRecipe`
            An ArmyRecipe object with the following attributes:
            
            - ``heroes_loadout``: List[:class:`HeroLoadout`] - Hero loadouts with pets and equipment
            - ``troops``: List[Tuple[:class:`Troop`, int]] - Troops with their quantities
            - ``spells``: List[Tuple[:class:`Spell`, int]] - Spells with their quantities
            - ``clan_castle_troops``: List[Tuple[:class:`Troop`, int]] - Clan castle troops with quantities
            - ``clan_castle_spells``: List[Tuple[:class:`Spell`, int]] - Clan castle spells with quantities

        """
        return ArmyRecipe(static_data=self._static_data, link=link)

    def parse_account_data(self, data: dict) -> AccountData:
        """Parse account data JSON from in-game into an AccountData object.

        Parameters
        ----------
        data: dict
            The raw account data.

        Returns
        --------
        :class:`AccountData`
            An AccountData object containing parsed account information.

        """
        return AccountData(data=data, client=self)

    def get_troop(
        self, name: str, is_home_village: bool = True, level: int = 0
    ) -> Optional["Troop"]:
        """Get a Troop object with the given name and level.

        Example
        -------

        .. code-block:: python3

            troop = client.get_troop("Barbarian", level=5)
            print(f"{troop.name} has {troop.dps} DPS at Lv{troop.level}")


        Parameters
        ----------
        name: str
            The troop name, which must match in-game **exactly**, and is case-sensitive.

        is_home_village: bool
            Whether the troop belongs to the home village or builder base. Defaults to True (home village).

        level: int
            The level of the troop. Defaults to 1.

        Returns
        --------
        Optional[:class:`Troop`]
            A Troop object at the specified level, or ``None`` if the troop is not found.

        """

        troop_data = self._get_static_data(
            item_name=name,
            section="troops",
            village="home" if is_home_village else "builderBase"
        )
        if troop_data is None:
            return None
        return Troop(data={}, static_data=troop_data, level=level)

    def get_spell(self, name: str, level: int = 0) -> Optional["Spell"]:
        """Get a Spell object with the given name and level.

        Example
        -------

        .. code-block:: python3

            spell = client.get_spell("Healing Spell", level=3)
            print(f"{spell.name} has {spell.upgrade_cost} upgrade cost at Lv{spell.level}")


        Parameters
        ----------
        name: str
            The spell name, which must match in-game **exactly**, and is case-sensitive.

        level: int
            The level of the spell. Defaults to 1.

        Returns
        --------
        Optional[:class:`Spell`]
            A Spell object at the specified level, or ``None`` if the spell is not found.

        """

        spell_data = self._get_static_data(
            item_name=name,
            section="spells",
        )
        if spell_data is None:
            return None
        return Spell(data={}, static_data=spell_data, level=level)

    def get_hero(self, name: str, level: int = 0) -> Optional["Hero"]:
        """Get a Hero object with the given name and level.

        Example
        -------

        .. code-block:: python3

            hero = client.get_hero("Archer Queen", level=50)
            print(f"{hero.name} has {hero.upgrade_cost} upgrade cost at Lv{hero.level}")


        Parameters
        ----------
        name: str
            The hero name, which must match in-game **exactly**, and is case-sensitive.

        level: int
            The level of the hero. Defaults to 1.

        Returns
        --------
        Optional[:class:`Hero`]
            A Hero object at the specified level, or ``None`` if the hero is not found.

        """

        hero_data = self._get_static_data(
            item_name=name,
            section="heroes",
        )
        if hero_data is None:
            return None
        return Hero(data={}, static_data=hero_data, level=level)

    def get_pet(self, name: str, level: int = 0) -> Optional["Pet"]:
        """Get a Pet object with the given name and level.

        Example
        -------

        .. code-block:: python3

            pet = client.get_pet("Electro Owl", level=5)
            print(f"{pet.name} has {pet.upgrade_cost} upgrade cost at Lv{pet.level}")


        Parameters
        ----------
        name: str
            The pet name, which must match in-game **exactly**, ahd is case-sensitive.

        level: int
            The level of the pet. Defaults to 1.

        Returns
        --------
        Optional[:class:`Pet`]
            A Pet object at the specified level, or ``None`` if the pet is not found.

        """
        pet_data = self._get_static_data(
            item_name=name,
            section="pets",
        )
        if pet_data is None:
            return None
        return Pet(data={}, static_data=pet_data, level=level)

    def get_equipment(self, name: str, level: int = 0) -> Optional["Equipment"]:
        """Get an Equipment object with the given name and level.

        Example
        -------

        .. code-block:: python3

            gear = client.get_equipment("Dark Orb", level=10)
            print(f"{gear.name} has {gear.upgrade_cost} upgrade cost at Lv{gear.level}")


        Parameters
        ----------
        name: str
            The equipment name, which must match in-game **exactly**, and is case-sensitive.

        level: int
            The level of the equipment. Defaults to 1.

        Returns
        --------
        Optional[:class:`Equipment`]
            An Equipment object at the specified level, or ``None`` if the equipment is not found.

        """
        equipment_data = self._get_static_data(
            item_name=name,
            section="equipment",
        )
        if equipment_data is None:
            return None
        return Equipment(data={}, static_data=equipment_data, level=level)

    def get_translation(self, translation_id: str) -> Optional[Translation]:
        """Get translation data for a given Translation ID (TID).

        Example
        -------

        .. code-block:: python3

            translation = client.get_translation("TID_HERO_PALADIN_CHAMPION")
            if translation:
                print(f"English: {translation.english}")
                print(f"Russian: {translation.russian}")
                print(f"By code: {translation['RU']}")


        Parameters
        ----------
        TID: str
            The Translation ID string (e.g., "TID_HERO_PALADIN_CHAMPION").

        Returns
        --------
        :class:`Translation` | None
            A Translation object with all language translations, or ``None`` if the TID is not found.

        """
        translation_data: dict | None = self._translations.get(translation_id)
        if not translation_data:
            return None
        return Translation(data=translation_data)

    def get_extended_cwl_group_data(self, name: str) -> Optional[ExtendedCWLGroup]:
        """Get extended CWL group data by league name.

        Example
        -------

        .. code-block:: python3

            cwl_group = client.get_extended_cwl_group_data("Champion League I")
            print(f"{cwl_group.name}: {cwl_group.first_place_medals} medals for 1st place")


        Parameters
        ----------
        name: str
            The CWL league name, which must match in-game **exactly**, and is case-sensitive.

        Returns
        --------
        Optional[:class:`ExtendedCWLGroup`]
            An ExtendedCWLGroup object containing CWL league information, or ``None`` if not found.

        """
        cwl_data: Optional[dict] = self._get_static_data(
            item_name=name,
            section="war_leagues",
        )
        if not cwl_data:
            return None
        return ExtendedCWLGroup(data=cwl_data)



# Enables circular import for type hinting coc.Client
from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from typing import Optional, TYPE_CHECKING, Type, Union

from .raid import RaidLogEntry
from .wars import ClanWarLogEntry

if TYPE_CHECKING:
    from .client import Client


class LogPaginator(ABC):
    @abstractmethod
    def __init__(self, client: Client,
                 clan_tag: str,
                 limit: int,
                 page: bool,
                 json_resp: dict,
                 model: Union[Type[ClanWarLogEntry], Type[RaidLogEntry]],
                 **kwargs):

        self._clan_tag = clan_tag
        self._limit = limit
        self._page = page
        
        self.kwargs = kwargs
        self.kwargs["lookup_cache"] = kwargs.get("lookup_cache", client.lookup_cache if client else None)
        self.kwargs["update_cache"] = kwargs.get("update_cache", client.update_cache if client else None)
        self.kwargs["ignore_cached_errors"] = kwargs.get("ignore_cached_errors", client.ignore_cached_errors if client else None)

        self._init_data = json_resp  # Initial data; this is const
        self._init_logs = json_resp.get("items", [])
        self._response_retry = json_resp.get("_response_retry", 0)
        self._client = client
        self._model = model

    def __len__(self) -> int:
        return len(self._init_logs)

    def __iter__(self):
        """Initialize the iter object and reset the iter index to 0"""
        self._sync_index = 0
        return self

    def __next__(self) -> Union[ClanWarLogEntry, RaidLogEntry]:
        """Fetch the next item in the iter object and return the entry"""
        if self._sync_index == len(self._init_logs):
            raise StopIteration
        ret = self._model(data=self._init_logs[self._sync_index],
                          client=self._client, response_retry=self._response_retry, clan_tag=self._clan_tag)
        self._sync_index += 1
        return ret

    def __getitem__(self, index: int) -> Union[ClanWarLogEntry, RaidLogEntry]:
        """Support indexing the object. This will not fetch any addition
        items from the endpoint"""
        try:
            return self._model(data=self._init_logs[index],
                               client=self._client, response_retry=self._response_retry, clan_tag=self._clan_tag)
        except Exception:
            raise

    def __aiter__(self):
        # These values are used to simulate the caller having a single list
        # of items. In reality, the list is populated on demand.
        self._max_index = len(self._init_logs)
        self._min_index = 0
        self._async_index = 0

        # Make copies of the init data since they will change.
        self._logs = self._init_logs[:]
        self._page_data = self._init_data.copy()
        return self

    async def __anext__(self) -> Union[ClanWarLogEntry, RaidLogEntry]:
        """
        This class supports async for loops. If the `page` bool is set to
        True then the async for loop will fetch all items from the endpoint
        until there are not more items in the endpoint. This is done without
        increasing the memory footprint by only caching `limit` number
        of logs at all times.

        When `limit` is set to 10, `self._logs` will only store 10 log items.
        When the last item in `self._logs` is reached when iterating,
        the array will be replaced by the next `limit` number of items. All
        this is abstracted from the user, they will just think they are
        iterating over the array. Keep in mind that if `limit` is set to 10
        and there are 200 total logs, then this API will make 20 get
        requests to the endpoint making it quite slow. Consider tuning
        this method with the `limit` value.
        """
        # If paging is not enabled, do not fetch  any more items only
        # iterate over the items in the self._war_logs
        if not self._page:
            if self._async_index == len(self._logs):
                raise StopAsyncIteration
            ret = self._model(data=self._logs[self._async_index],
                              client=self._client, response_retry=self._response_retry, clan_tag=self._clan_tag)
            self._async_index += 1
            return ret

        # If paging is enabled, update self._war_logs if the end of the
        # array is reached
        ret: Union[ClanWarLogEntry, RaidLogEntry]

        # If index request is within range of the war_logs, return item
        if self._min_index <= self._async_index < self._max_index:
            ret = self._logs[self._async_index - self._min_index]

        # Iteration has reached the end of the array, fetch the next
        # set of logs from the endpoint
        elif self._next_page:
            await self._paginate()
            self._min_index = self._max_index
            self._max_index = self._max_index + len(self._logs)
            ret = self._logs[self._async_index - self._min_index]
        else:
            raise StopAsyncIteration

        self._async_index += 1
        return self._model(data=ret, client=self._client, response_retry=self._response_retry, clan_tag=self._clan_tag)

    async def _paginate(self) -> None:
        """
        Request data from the endpoint and update the iter variables with
        the new data. `self._fetch_endpoint` is a child defined method.
        """
        self._page_data = await self._fetch_endpoint(self._client,
                                                     self._clan_tag,
                                                     **self.options)
        self._logs = self._page_data.get("items", [])
        self._response_retry = self._page_data.get("_response_retry", 0)

    @property
    def options(self) -> dict:
        """Generate the header for the endpint request"""
        options = {"limit": self._limit, **self.kwargs}
        if self._next_page:
            options["after"] = self._next_page
        return options

    @property
    def _next_page(self) -> Optional[str]:
        """Determine if there is a next page for the endpoint query"""
        try:
            return self._page_data.get("paging").get("cursors").get("after")
        except KeyError:
            return None

    @staticmethod
    @abstractmethod
    async def _fetch_endpoint(client: Client,
                              clan_tag: str,
                              fut: Optional[asyncio.Future] = None,
                              **options) -> dict:
        """Function to fetch data from the endpoint"""
        pass

    @classmethod
    @abstractmethod
    async def init_cls(cls,
                       client: Client,
                       clan_tag: str,
                       model: Type[ClanWarLogEntry],
                       limit: int,
                       paginate: bool = True,
                       **kwargs,
                       ) -> Union[ClanWarLog, RaidLog]:
        """Class method to return an instantiated object"""
        pass


class ClanWarLog(LogPaginator, ABC):
    """Represents a Generator for a ClanWarLog"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @classmethod
    async def init_cls(cls,
                       client: Client,
                       clan_tag: str,
                       model: Type[ClanWarLogEntry],
                       limit: int,
                       page: bool = True,
                       after: str = None,
                       before: str = None,
                       **kwargs
                       ) -> ClanWarLog:

        # Add the limit if specified
        args = {"limit": limit} if limit else {}
        if after:
            args["after"] = after
        if before:
            args["before"] = before
        args["lookup_cache"] = kwargs.get("lookup_cache", client.lookup_cache)
        args["update_cache"] = kwargs.get("update_cache", client.update_cache)
        args["ignore_cached_errors"] = kwargs.get("ignore_cached_errors", client.ignore_cached_errors)

        json_resp = await cls._fetch_endpoint(client, clan_tag, **args)
        return ClanWarLog(client=client, clan_tag=clan_tag, limit=limit,
                          page=page, json_resp=json_resp, model=model, **kwargs)

    @staticmethod
    async def _fetch_endpoint(client: Client, clan_tag: str,
                              fut: Optional[asyncio.Future] = None,
                              **options) -> dict:
        result = await client.http.get_clan_war_log(clan_tag, **options)
        if fut:
            fut.set_result(result)
        return result


class RaidLog(LogPaginator, ABC):
    """Represents a Generator for a RaidLog"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @classmethod
    async def init_cls(cls,
                       client: Client,
                       clan_tag: str,
                       model: Type[RaidLogEntry],
                       limit: int,
                       page: bool = True,
                       after: str = None,
                       before: str = None,
                       **kwargs
                       ) -> RaidLog:

        # Add the limit if specified
        args = {"limit": limit} if limit else {}
        if after:
            args["after"] = after
        if before:
            args["before"] = before
        args["lookup_cache"] = kwargs.get("lookup_cache", client.lookup_cache)
        args["update_cache"] = kwargs.get("update_cache", client.update_cache)
        args["ignore_cached_errors"] = kwargs.get("ignore_cached_errors", client.ignore_cached_errors)

        json_resp = await cls._fetch_endpoint(client, clan_tag, **args)
        return RaidLog(client=client, clan_tag=clan_tag, limit=limit,
                       page=page, json_resp=json_resp, model=model, **kwargs,)

    @staticmethod
    async def _fetch_endpoint(client: Client, clan_tag: str,
                              fut: Optional[asyncio.Future] = None,
                              **options) -> dict:
        result = await client.http.get_clan_raid_log(clan_tag, **options)
        if fut:
            fut.set_result(result)
        return result



"""An extension that provides decorators to facilitate automated, periodic repetition of functions."""


from .cron import CronSchedule, CronParserError
from .triggers import BaseTrigger, CronTrigger, IntervalTrigger, on_error, start_triggers

__all__ = [
    'BaseTrigger',
    'CronParserError',
    'CronSchedule',
    'CronTrigger',
    'IntervalTrigger',
    'on_error',
    'start_triggers'
]




import asyncio
import logging
import re

from collections import deque
from datetime import datetime, timezone
from itertools import cycle
from time import process_time, perf_counter
from typing import Optional
from urllib.parse import urlencode
from base64 import b64decode as base64_b64decode
from json import loads as json_loads

import aiohttp
import orjson

from .errors import (
    HTTPException,
    Maintenance,
    NotFound,
    InvalidArgument,
    Forbidden,
    InvalidCredentials,
    GatewayError,
)
from .utils import FIFO, HTTPStats

LOG = logging.getLogger(__name__)
KEY_MINIMUM, KEY_MAXIMUM = 1, 10
stats_url_matcher = re.compile(r"%23[\da-zA-Z]+|\d{8,}|global")


async def json_or_text(response: aiohttp.ClientResponse):
    """Parses an aiohttp response into a the string or json response."""
    try:
        ret = await response.json(loads=orjson.loads)
    except aiohttp.ContentTypeError:
        ret = await response.text(encoding="utf-8")

    return ret


class BasicThrottler:
    """Basic throttler that sleeps for `sleep_time` seconds between each request."""

    __slots__ = (
        "sleep_time",
        "last_run",
        "lock",
    )

    def __init__(self, sleep_time):
        self.sleep_time = sleep_time
        self.last_run = None
        self.lock = asyncio.Lock()
        LOG.debug("BasicThrottler initialized with sleeptime %s", self.sleep_time)

    async def __aenter__(self):
        async with self.lock:
            last_run = self.last_run
            if last_run:
                difference = process_time() - last_run
                need_to_sleep = self.sleep_time - difference
                if need_to_sleep > 0:
                    LOG.debug("Request throttled. Sleeping for %s", need_to_sleep)
                    await asyncio.sleep(need_to_sleep)

            self.last_run = process_time()
            return self

    async def __aexit__(self, exception_type, exception, traceback):
        pass


class BatchThrottler:
    """Simple throttler that allows `rate_limit` requests (per second) before sleeping until the next second."""

    __slots__ = (
        "rate_limit",
        "per",
        "retry_interval",
        "_task_logs",
    )

    def __init__(self, rate_limit, per=1.0, retry_interval=0.001):
        self.rate_limit = rate_limit
        self.per = per
        self.retry_interval = retry_interval

        self._task_logs = deque()
        LOG.debug("BatchThrottler initialized with rate_limit %s, per %s, retry_interval %s", self.rate_limit, self.per,
                  self. retry_interval)

    async def __aenter__(self):
        while True:
            now = process_time()

            # Pop items(which are start times) that are no longer in the
            # time window
            while self._task_logs:
                if now - self._task_logs[0] > self.per:
                    self._task_logs.popleft()
                else:
                    break

            # Exit the infinite loop when new task can be processed
            if len(self._task_logs) < self.rate_limit:
                break

            retry_interval = self.retry_interval
            LOG.debug("Request throttled. Sleeping for %s seconds.", retry_interval)
            await asyncio.sleep(retry_interval)

        # Push new task's start time
        self._task_logs.append(process_time())

        return self

    async def __aexit__(self, exception_type, exception, traceback):
        pass


class Route:
    """Helper class to create endpoint URLs."""
    ignored_kwargs = ['lookup_cache', 'update_cache', 'ignore_cached_errors']

    def __init__(self, method: str, base: str, path: str, **kwargs: dict):
        """
        The class is used to create the final URL used to fetch the data
        from the API. The parameters that are passed to the API are all in
        the GET request packet. This class will parse the `kwargs` dictionary
        and concatenate any parameters passed in.

        Parameters
        ----------
        method:
            :class:`str`: HTTP method used for the HTTP request
        base:
            :class:`str`: Base URL used for the HTTP request
        path:
            :class:`str`: URL path used for the HTTP request
        kwargs:
            :class:`dict`: Optional options used to concatenate into the final
            URL
        """
        if "#" in path:
            path = path.replace("#", "%23")

        self.method = method
        self.path = path
        self.base = base

        url = self.base + self.path

        if kwargs:
            url_params = {}
            for k, v in kwargs.items():
                if v is None:
                    continue
                if k in self.ignored_kwargs:
                    continue
                url_params[k] = v
            self.url = "{}?{}".format(url, urlencode(url_params))
        else:
            self.url = url

    @property
    def stats_key(self):
        return stats_url_matcher.sub("{}", self.path)


class HTTPClient:
    """HTTP Client for the library. All low-level requests and key-management occurs here."""

    # pylint: disable=too-many-arguments, missing-docstring, protected-access, too-many-branches
    def __init__(
            self,
            client,
            loop,
            email,
            password,
            key_names,
            key_count,
            key_scopes,
            throttle_limit,
            throttler=BasicThrottler,
            cache_max_size=10000,
            stats_max_size=1000,
            base_url="https://api.clashofclans.com/v1",
            ip=None,
            lookup_cache=True,
            update_cache=True,
            ignore_cached_errors=None,
    ):
        self.aiohttp_request_kwargs = ['params', 'data', 'json', 'cookies', 'headers', 'skip_auto_headers',
                                       'auth', 'allow_redirects', 'max_redirects', 'compress', 'chunked', 'expect100',
                                       'raise_for_status', 'read_until_eof', 'proxy', 'proxy_auth', 'timeout',
                                       'ssl', 'server_hostname', 'proxy_headers', 'trace_request_ctx', 'read_bufsize',
                                       'auto_decompress', 'max_line_size', 'max_field_size']
        self.client = client
        self.loop = loop
        self.email = email
        self.password = password
        self.key_names = key_names
        self.key_count = key_count
        self.key_scopes = key_scopes
        self.throttle_limit = throttle_limit
        per_second = key_count * throttle_limit
        self.lookup_cache = lookup_cache
        self.update_cache = update_cache
        self.ignore_cached_errors = ignore_cached_errors or []
        self.__session: Optional[aiohttp.ClientSession] = None
        self.__lock = asyncio.Semaphore(per_second)
        self.cache = cache_max_size and FIFO(cache_max_size)
        self._cache_remove_count = 0
        self.stats = stats_max_size and HTTPStats(max_size=stats_max_size)
        if base_url and isinstance(base_url, str) and len(base_url) > 0:
            if base_url.endswith("/"):
                base_url = base_url[:-1]
            self.base_url = base_url
        else:
            raise ValueError("base_url must be a string and not empty.")
        self.ip = ip
        if issubclass(throttler, BasicThrottler):
            self.__throttle = throttler(1 / per_second)
        elif issubclass(throttler, BatchThrottler):
            self.__throttle = throttler(per_second)
        else:
            raise TypeError("throttler must be either BasicThrottler or BatchThrottler.")

        self._keys = []
        self.keys = None

        self.initialising_keys = asyncio.Event()
        self.initialising_keys.set()

    def _cache_remove(self, key):
        try:
            del self.cache[key]
            #  The following fixes a memory leak that is caused by python dicts not properly freeing disk space
            self._cache_remove_count += 1
            if self._cache_remove_count >= self.cache.max_size:
                self.cache = self.cache.copy()
                self._cache_remove_count = 0
                LOG.debug("Cache copied to prevent a memory leak")
        except KeyError:
            pass

    async def create_session(self, connector, timeout):
        self.__session = aiohttp.ClientSession(connector=connector, timeout=aiohttp.ClientTimeout(total=timeout))

    async def close(self):
        if self.__session:
            await self.__session.close()

    async def request(self, route, **kwargs):
        method = route.method
        url = route.url

        headers = {
            "Accept"       : "application/json",
            "authorization": "Bearer {}".format(next(self.keys)),
            "Accept-Encoding": "gzip, deflate",
        }
        kwargs["headers"] = headers

        if "json" in kwargs:
            kwargs["headers"]["Content-Type"] = "application/json"

        cache_control_key = route.url
        cache = self.cache
        lookup_cache = kwargs.pop("lookup_cache", self.lookup_cache)
        update_cache = kwargs.pop("update_cache", self.update_cache)
        ignore_cached_errors = kwargs.pop("ignore_cached_errors", self.ignore_cached_errors)
        # the cache will be cleaned once it becomes stale / a new object is available from the api.
        if isinstance(cache, FIFO) and (lookup_cache or (lookup_cache is None and 'realtime' not in url)):
            try:
                data = cache[cache_control_key]
                status_code = data.get("status_code")
                if data.get("timestamp") and data.get("timestamp") + data.get("_response_retry", 0) < datetime.now(
                        tz=timezone.utc).timestamp():
                    self._cache_remove(cache_control_key)
                elif not status_code or 200 <= status_code < 300:
                    return data
                # ignore status cached errors if wanted
                elif isinstance(ignore_cached_errors, list) and status_code in ignore_cached_errors:
                    pass
                elif status_code == 400:
                    raise InvalidArgument(400, data)
                elif status_code == 403:
                    raise Forbidden(403, data)
                elif status_code == 404:
                    raise NotFound(404, data)
                elif status_code == 503:
                    raise Maintenance(503, data)
            except KeyError:
                pass
        request_kwargs = {k: v for k, v in kwargs.items() if k in self.aiohttp_request_kwargs}
        for tries in range(5):
            try:
                async with self.__lock, self.__throttle:
                    start = perf_counter()
                    async with self.__session.request(method, url, **request_kwargs) as response:

                        perf = (perf_counter() - start) * 1000
                        log_info = {"method": method, "url": url, "perf_counter": perf, "status": response.status}
                        if isinstance(self.stats, HTTPStats):
                            self.stats[route.stats_key] = perf

                        LOG.debug("API HTTP Request: %s", str(log_info))
                        data = (await json_or_text(response)) or {}
                        if isinstance(data, dict):
                            data["status_code"] = response.status
                            data["timestamp"] = datetime.now(tz=timezone.utc).timestamp()
                        try:
                            # set a callback to remove the item from cache once it's stale.
                            delta = int(response.headers["Cache-Control"].strip("max-age=").strip("public max-age="))
                            # encounter for changed description in cache control header. for realtime it is always
                            # 600 but that is not true. Correct is 0
                            data["_response_retry"] = delta if 'realtime' not in url else 0
                            if isinstance(cache, FIFO) and (update_cache or (update_cache is None and 'realtime' not in url)):
                                self.cache[cache_control_key] = data
                                LOG.debug("Cache-Control max age: %s seconds, key: %s", delta, cache_control_key)
                                self.loop.call_later(delta, self._cache_remove, cache_control_key)

                        except (KeyError, AttributeError, ValueError):
                            # the request didn't contain cache control headers so skip any cache handling.
                            # if the API returns a timeout error (504) it will return a string of HTML.
                            if isinstance(data, dict):
                                data["_response_retry"] = 0

                        if 200 <= response.status < 300:
                            LOG.debug("%s has received %s", url, data)
                            return data

                        if response.status == 400:
                            raise InvalidArgument(response, data)

                        if response.status == 403:
                            LOG.info("forbidden! resp: %s, msg: %s", str(response), str(data))
                            if data.get("reason") == "accessDenied.invalidIp" and self.email and self.password:
                                if self.initialising_keys.is_set():
                                    await self.initialise_keys()

                                await self.initialising_keys.wait()
                                return await self.request(route, **kwargs)

                            raise Forbidden(response, data)

                        if response.status == 404:
                            raise NotFound(response, data)
                        if response.status == 429:
                            LOG.error(
                                    "We have been rate-limited by the API. "
                                    "Reconsider the number of requests you are allowing per second."
                            )
                            raise HTTPException(response, data)

                        if response.status == 503:
                            if isinstance(data, str):
                                # weird case where a 503 will be raised, but html returned.
                                text = re.compile(r"<[^>]+>").sub(data, "")
                                raise Maintenance(response, text)

                            raise Maintenance(response, data)

                        if response.status in (500, 502, 504):
                            # gateway error, retry again
                            await asyncio.sleep(tries * 2 + 1)
                            continue

                        # catch any stray status codes
                        raise HTTPException(response, data)

            except asyncio.TimeoutError:
                # api timed out, retry again
                if tries > 3:
                    raise GatewayError("The API timed out waiting for the request.")

                await asyncio.sleep(tries * 2 + 1)
                continue

        else:
            if response.status in (500, 502, 504):
                if isinstance(data, str):
                    # gateway errors return HTML
                    text = re.compile(r"<[^>]+>").sub(data, "")
                    raise GatewayError(response, text)

                raise GatewayError(response, data)

            raise HTTPException(response, data)

    # clans

    def search_clans(self, **kwargs):
        return self.request(Route("GET", self.base_url, "/clans", **kwargs), **kwargs)

    def get_clan(self, tag, **kwargs):
        return self.request(Route("GET", self.base_url, "/clans/{}".format(tag)), **kwargs)

    def get_clan_members(self, tag, **kwargs):
        return self.request(Route("GET", self.base_url, "/clans/{}/members".format(tag), **kwargs), **kwargs)

    def get_clan_war_log(self, tag, **kwargs):
        return self.request(Route("GET", self.base_url, "/clans/{}/warlog".format(tag), **kwargs), **kwargs)

    def get_clan_current_war(self, tag, realtime=None, **kwargs):
        return self.request(Route("GET", self.base_url, "/clans/{}/currentwar".format(tag) + (
                                         '?realtime=true' if realtime or (realtime is None and self.client.realtime)
                                         else '')), **kwargs)

    def get_clan_war_league_group(self, tag, realtime=None, **kwargs):
        return self.request(Route("GET", self.base_url, "/clans/{}/currentwar/leaguegroup".format(tag) + (
                                         '?realtime=true' if realtime or (realtime is None and self.client.realtime)
                                         else '')), **kwargs)

    def get_cwl_wars(self, war_tag, realtime=None, **kwargs):
        return self.request(Route("GET", self.base_url, "/clanwarleagues/wars/{}".format(war_tag) + (
                                         '?realtime=true' if realtime or (realtime is None and self.client.realtime)
                                         else '')), **kwargs)

    def get_clan_raid_log(self, tag, **kwargs):
        return self.request(Route("GET", self.base_url, "/clans/{}/capitalraidseasons".format(tag), **kwargs), **kwargs)

    # locations

    def search_locations(self, **kwargs):
        return self.request(Route("GET", self.base_url, "/locations", **kwargs), **kwargs)

    def get_location(self, location_id, **kwargs):
        return self.request(Route("GET", self.base_url, "/locations/{}".format(location_id)), **kwargs)

    def get_location_clans(self, location_id, **kwargs):
        return self.request(Route("GET", self.base_url, "/locations/{}/rankings/clans".format(location_id), **kwargs), **kwargs)

    def get_location_players(self, location_id, **kwargs):
        return self.request(Route("GET", self.base_url, "/locations/{}/rankings/players".format(location_id), **kwargs), **kwargs)

    def get_location_clans_builder_base(self, location_id, **kwargs):
        return self.request(Route("GET", self.base_url, "/locations/{}/rankings/clans-builder-base".format(location_id), **kwargs),
                            **kwargs)

    def get_location_clans_capital(self, location_id, **kwargs):
        return self.request(Route("GET", self.base_url, "/locations/{}/rankings/capitals".format(location_id), **kwargs), **kwargs)

    def get_location_players_builder_base(self, location_id, **kwargs):
        return self.request(Route("GET", self.base_url, "/locations/{}/rankings/players-builder-base".format(location_id), **kwargs),
                            **kwargs)

    # leagues

    def search_leagues(self, **kwargs):
        return self.request(Route("GET", self.base_url, "/leaguetiers", **kwargs), **kwargs)

    def search_capital_leagues(self, **kwargs):
        return self.request(Route("GET", self.base_url, "/capitalleagues", **kwargs), **kwargs)

    def search_war_leagues(self, **kwargs):
        return self.request(Route("GET", self.base_url, "/warleagues", **kwargs), **kwargs)

    def search_builder_base_leagues(self, **kwargs):
        return self.request(Route("GET", self.base_url, "/builderbaseleagues", **kwargs), **kwargs)

    def get_league(self, league_id, **kwargs):
        return self.request(Route("GET", self.base_url, "/leaguetiers/{}".format(league_id)), **kwargs)

    def get_capital_league(self, league_id, **kwargs):
        return self.request(Route("GET", self.base_url, "/capitalleagues/{}".format(league_id)), **kwargs)

    def get_war_league(self, league_id, **kwargs):
        return self.request(Route("GET", self.base_url, "/warleagues/{}".format(league_id)), **kwargs)

    def get_builder_base_league(self, league_id, **kwargs):
        return self.request(Route("GET", self.base_url, "/builderbaseleagues/{}".format(league_id)), **kwargs)

    def get_league_seasons(self, league_id, **kwargs):
        return self.request(Route("GET", self.base_url, "/leagues/{}/seasons".format(league_id), **kwargs), **kwargs)

    def get_league_season_info(self, league_id, season_id, **kwargs):
        return self.request(Route("GET", self.base_url, "/leagues/{}/seasons/{}".format(league_id, season_id), **kwargs),
                            **kwargs)

    # players

    def get_player(self, player_tag, **kwargs):
        return self.request(Route("GET", self.base_url, "/players/{}".format(player_tag)), **kwargs)

    def verify_player_token(self, player_tag, token, **kwargs):
        return self.request(Route("POST", self.base_url, "/players/{}/verifytoken".format(player_tag)),
                            json={"token": token}, **kwargs)

    # labels

    def get_clan_labels(self, **kwargs):
        return self.request(Route("GET", self.base_url, "/labels/clan", **kwargs), **kwargs)

    def get_player_labels(self, **kwargs):
        return self.request(Route("GET", self.base_url, "/labels/players", **kwargs), **kwargs)

    def get_current_goldpass_season(self, **kwargs):
        return self.request(Route("GET", self.base_url, "/goldpass/seasons/current"), **kwargs)

    # key updating management

    async def initialise_keys(self):
        LOG.debug("Initialising keys from the developer site.")
        self.initialising_keys.clear()

        # Use context manager to automatically clean up after ourselves
        async with aiohttp.ClientSession() as session:
            body = {"email": self.email, "password": self.password}
            resp = await session.post("https://developer.clashofclans.com/api/login", json=body)
            if resp.status == 403:
                LOG.error("Invalid credentials used when attempting to log in")
                await self.close()
                raise InvalidCredentials()

            LOG.info("Successfully logged into the developer site.")

            resp_payload = await resp.json()
            if not self.ip:
                ip = json_loads(base64_b64decode(resp_payload["temporaryAPIToken"].split(".")[1] + "====").decode("utf-8"))["limits"][1]["cidrs"][0].split("/")[0]
            else:
                ip = self.ip
            LOG.info("Found IP address to be %s", ip)

            resp = await session.post("https://developer.clashofclans.com/api/apikey/list")
            keys = (await resp.json()).get("keys",{})
            for key in keys:
                LOG.debug(f"Key {key}")
                if key["name"] != self.key_names or ip not in key["cidrRanges"]:
                    continue
                self._keys.append(key["key"])
                if len(self._keys) == self.key_count:
                    break

            LOG.info("Retrieved %s valid keys from the developer site.", len(self._keys))

            if len(self._keys) < self.key_count:
                for key in keys[:]:
                    if key["name"] != self.key_names or ip in key["cidrRanges"]:
                        continue
                    LOG.info(
                            "Deleting key with the name %s and IP %s (not matching our current IP address).",
                            self.key_names, key["cidrRanges"],
                    )
                    resp = await session.post("https://developer.clashofclans.com/api/apikey/revoke", json={"id": key["id"]})
                    if resp.status == 200:
                        keys.remove(key)

                while len(self._keys) < self.key_count and len(keys) < KEY_MAXIMUM:
                    data = {
                        "name"       : self.key_names,
                        "description": "Created on {}".format(datetime.now().strftime("%c")),
                        "cidrRanges" : [ip],
                        "scopes"     : [self.key_scopes],
                    }

                    LOG.info("Creating key with data %s.", str(data))

                    resp = await session.post("https://developer.clashofclans.com/api/apikey/create", json=data)
                    key = await resp.json()

                    if resp.status != 200:
                        LOG.error(key.get("description"))
                        raise ValueError(key.get("description"))

                    self._keys.append(key["key"]["key"])

            if len(keys) == 10 and len(self._keys) < self.key_count:
                LOG.critical("%s keys were requested to be used, but a maximum of %s could be "
                             "found/made on the developer site, as it has a maximum of 10 keys per account. "
                             "Please delete some keys or lower your `key_count` level."
                             "I will use %s keys for the life of this client.",
                             self.key_count, len(self._keys), len(self._keys))

            if len(self._keys) == 0:
                await self.close()
                raise RuntimeError(
                        "There are {} API keys already created and none match a key_name of '{}'."
                        "Please specify a key_name kwarg, or go to 'https://developer.clashofclans.com' to delete "
                        "unused keys.".format(len(keys), self.key_names)
                )

        self.keys = cycle(self._keys)
        self.initialising_keys.set()
        LOG.info("Successfully initialised keys for use.")

    async def get_data_from_url(self, url):
        async with self.__session.get(url) as response:
            if response.status == 200:
                return await response.read()
            if response.status == 404:
                raise NotFound(response, "image not found")

            raise HTTPException(response, "failed to get image")



"""An extension that helps interact with the Clash of Clans Discord Junkies' Discord Links API."""

import asyncio
import base64
import logging
import typing
import json

from collections import namedtuple
from datetime import datetime, timezone

import aiohttp

from coc.http import json_or_text
from coc.utils import correct_tag

LOG = logging.getLogger(__name__)

AccessToken = namedtuple("AccessToken", ["token", "expires_at"])


def extract_expiry_from_jwt_token(token):
    if isinstance(token, str):
        token = token.encode("utf-8")
    elif not isinstance(token, bytes):
        # token was wrong somehow
        return None

    try:
        signing, _ = token.rsplit(b".", 1)
        _, payload = signing.split(b".", 1)
    except ValueError:
        return None  # not enough segments

    if len(payload) % 4 > 0:
        payload += b"=" * (4 - len(payload) % 4)

    bytes_payload = base64.urlsafe_b64decode(payload)
    dict_payload = json.loads(bytes_payload)
    try:
        expiry = dict_payload["exp"]
        return datetime.fromtimestamp(expiry, tz=timezone.utc)
    except KeyError:
        return None


async def login(username: str, password: str) -> "DiscordLinkClient":
    """Eases logging into the API client.

    For more information on this project, please join the discord server - <discord.gg/Eaja7gJ>

    You must have your username and password as given on the server.
    If unsure as to what this means, please reach out to an admin.

    Parameters
    -----------
    username : str
        Your username as given on the discord server.
    password : str
        Your password as given on the discord server
    loop : Optional[:class:`asyncio.AbstractEventLoop`]
        The :class:`asyncio.AbstractEventLoop` to use for HTTP requests.
        An :func:`asyncio.get_event_loop()` will be used if ``None`` is passed
    """
    if not isinstance(username, str) or not isinstance(password, str):
        raise TypeError("username and password must both be a string")
    if not username or not password:
        raise ValueError("username or password must not be an empty string.")

    loop = asyncio.get_running_loop()
    return DiscordLinkClient(username, password, loop)


class DiscordLinkClient:
    """An extension that helps interact with the Clash of Clans Discord Junkies' Discord Links API.

    For more information on this project, please join the discord server - <discord.gg/Eaja7gJ>

    You must have your username and password as given on the server.
    If unsure as to what this means, please reach out to an admin.

    Parameters
    -----------
    username : str
        Your username as given on the discord server.
    password : str
        Your password as given on the discord server

    loop : Optional[:class:`asyncio.AbstractEventLoop`]
        The :class:`asyncio.AbstractEventLoop` to use for HTTP requests.
        An :func:`asyncio.get_event_loop()` will be used if ``None`` is passed

    """

    BASE_URL = "https://cocdiscord.link"

    __slots__ = ("username", "password", "loop", "key", "http_session")

    def __init__(self, username: str, password: str, loop: asyncio.AbstractEventLoop = None):
        self.username = username
        self.password = password

        self.loop = loop or asyncio.get_event_loop()
        self.key = None  # set in get_key()

        self.http_session = aiohttp.ClientSession(loop=self.loop)

    async def close(self):
        """Close the client session established"""
        await self.http_session.close()

    async def _request(self, method, url, *, token_request: bool = False, **kwargs):
        url = self.BASE_URL + url

        if not token_request:
            key = await self._get_key()

            headers = {"authorization": "Bearer {}".format(key)}
            kwargs["headers"] = headers

        async with self.http_session.request(method, url, **kwargs) as response:
            LOG.debug("%s (%s) has returned %s", url, method, response.status)
            data = await json_or_text(response)
            LOG.debug(data)

            if 200 <= response.status < 300:
                LOG.debug("%s has received %s", url, data)
                return data

            if response.status == 401:
                await self._refresh_key()
                return await self._request(method, url, **kwargs)

    async def _get_key(self):
        if not self.key or (self.key.expires_at and (self.key.expires_at < datetime.now(tz=timezone.utc))):
            await self._refresh_key()

        return self.key.token

    async def _refresh_key(self):
        data = {
            "username": self.username,
            "password": self.password,
        }

        payload = await self._request("POST", "/login", token_request=True, json=data)

        self.key = AccessToken(payload["token"], extract_expiry_from_jwt_token(payload["token"]))

    async def get_link(self, player_tag: str) -> typing.Optional[int]:
        """Get a linked discord ID of a player tag.
        Player tags can be found either in game or by from clan member lists.

        Parameters
        ----------
        player_tag: str
            The player tag to search for.

        Returns
        --------
        Optional[:class:`int`]
            The discord ID linked to the player, or ``None`` if no link found.
        """
        data = await self._request("GET", "/links/{}".format(correct_tag(player_tag, prefix="")))
        try:
            return int(data[0]["discordId"])
        except (IndexError, KeyError, TypeError, ValueError):
            return None

    async def get_links(self, *player_tag: str) -> typing.List[typing.Tuple[str, typing.Optional[int]]]:
        r"""Get linked discord IDs for an iterable of player tags.
        Player tags can be found either in game or by from clan member lists.

        This is the recommended method to use when fetching links for multiple tags as it uses a different endpoint.

        Parameters
        ----------
        \*player_tag: :class:`str`
            The player tags to search for.

        Returns
        --------
        List[:class:`str`:, Optional[:class:`int`]]
            A list of player_tag, discord_id tuple matches. Discord ID will be ``None`` if not found.

        Example
        --------

        .. code-block:: python3

            links = await client.get_links("#tag1", "#tag2", "#tag3")

            for player_tag, discord_id in links:
                print(player_tag, discord_id)

        """
        tags = [correct_tag(tag, prefix="") for tag in player_tag]
        data = await self._request("POST", "/batch", json=tags)
        data = data or []

        unclaimed_tags = set("#" + tag for tag in tags) - set(p["playerTag"] for p in data)

        return [(p["playerTag"], int(p["discordId"])) for p in data] + [(tag, None) for tag in unclaimed_tags]

    async def get_linked_players(self, discord_id: int) -> typing.List[str]:
        """Get a list of player tags linked to a discord ID.

        Parameters
        ----------
        discord_id: int
            The discord ID to search for.

        Returns
        --------
        List[:class:`str`]
         A list of player tags attached to the discord ID. If no links found, this will be an empty list.
        """
        data = await self._request("GET", "/links/{}".format(discord_id))
        if not data:
            return []

        return [item["playerTag"] for item in data]

    async def get_many_linked_players(self, *discord_id: int) -> typing.List[typing.Tuple[str, int]]:
        r"""Get a linked discord ID of a player tag.

        This is the recommended method to use when fetching links for multiple IDs as it uses a different endpoint.

        Parameters
        -----------
        \*discord_id: :class:`str`
            The discord IDs to search for.

        Returns
        --------
        List[Tuple[:class:`int`, :class:`str`]]
            A list containing (discord_id, tag) matches.

        Example
        -------

        .. code-block:: python3

            links = await client.get_many_linked_players(123456789, 234567890, 345678901)

            for player_tag, discord_id in links:
                print("{} is linked to {}".format(discord_id, player_tag))
        """
        data = await self._request("POST", "/batch", json=[str(n) for n in discord_id])
        if not data:
            return []

        return [(n["playerTag"], int(n["discordId"])) for n in data]

    async def add_link(self, player_tag: str, discord_id: int):
        """Creates a link between a player tag and a discord ID for the shared junkies database.
        Player tags can be found either in game or by from clan member lists.

        Parameters
        ----------
        player_tag : str
            The player tag to add the link to.
        discord_id: int
            The discord ID to add the link to.
        """
        data = {"playerTag": correct_tag(player_tag, prefix=""), "discordId": str(discord_id)}
        return await self._request("POST", "/links", json=data)

    async def delete_link(self, player_tag: str):
        """Deletes a link between a player tag and a discord ID for the shared junkies database.

       Player tags can be found either in game or by from clan member lists.

       Parameters
       ----------
       player_tag : str
           The player tag to remove the link from.
       """
        return await self._request("DELETE", "/links/{}".format(correct_tag(player_tag, prefix="")))
